#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author    : Yang
# Creadted  : 2017-04-17 23:10:28
from __future__ import print_function
import os
import sys
import time
import subprocess
import signal
import psutil
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

def log(s):
    print("[Monitor] %s" % s)

class PyChangeEventHandler(FileSystemEventHandler):
    def __init__(self, handler):
        super(PyChangeEventHandler, self).__init__()
        self.handler = handler
        self.last_handler = None

    def on_any_event(self, event):
        if self.last_handler and self.last_handler + 3 >= time.time():
            return
        if event.src_path.endswith(".go"):
            log("Go source file changed: %s" % event.src_path)
            self.handler()
            self.last_handler = time.time()


class WatchGo(object):
    def __init__(self, command, path, recursive=True):
        self.command = command
        self.path = path
        self.recursive = recursive
        self.process = None

    def start_process(self):
        log("Start process [%s]..." % ' '.join(self.command))
        self.process = subprocess.Popen(self.command, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
        log("Pid: %s" % self.process.pid)

    def kill_process(self):
        if self.process:
            log("Kill process [%s]..." % self.process.pid)
            try:
                parent = psutil.Process(self.process.pid)
            except psutil.NoSuchProcess:
                return
            children = parent.children(recursive=True)
            for pc in children:
                pc.send_signal(signal.SIGTERM)
            self.process = None

    def restart_process(self):
        self.kill_process()
        self.start_process()

    def start_watch(self):
        observer = Observer()
        observer.schedule(PyChangeEventHandler(self.restart_process), self.path, recursive=self.recursive)
        observer.start()
        log("Wathing directory %s..." % self.path)
        self.start_process()
        try:
            while True:
                time.sleep(0.5)
                if self.process.poll() == 0:
                    observer.stop()
                    break
        except KeyboardInterrupt:
            observer.stop()
        observer.join()




#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2020-09-11
import os

from .conf import *
from .helper import *

CMD_TPL = '''
package cmd

import (
	"flag"
	"fmt"
)

func init() {{
	appcmd.AddCmd("{name}", "{desc}", {name}Cmd)
}}

func {name}Cmd(args []string) (err error) {{
	var (
        {vars}
	)

    {params}

	err = param.Parse(args)
	if err != nil {{
		return
	}}

	if len(args) == 0 {{ // show help
		param.Usage()
		return
	}}

    fmt.Println("cmd: {name} running...")
	// write your code blow

	return
}}
'''


# Api model to code
'''
root cmd
CmdName # comment as cmd description
    param0 fieldType <defaultValue> # comment as param description, fieldType in [bool, str, int, float], option default value
    param1 fieldType <defaultValue> # comment as param description, fieldType in [bool, str, int, float], option default value
    param2 fieldType <defaultValue> # comment as param description, fieldType in [bool, str, int, float], option default value
'''

class CmdParam(object):
    def __init__(self, lineno, line, name, typ, defval, comment):
        self.lineno = lineno
        self.line = line
        self.name = name
        self.typ = typ
        self.defval = defval
        self.comment = comment

    @classmethod
    def fromLine(cls, lineno, line):
        if '#' in line:
            itemsline, comment = line.split('#', 1)
            items = itemsline.split()
        else:
            comment = ''
            items = line.split()
        assert_error(len(items) >= 2, '无效的参数定义', lineno, line)
        name, typ = items[0], items[1]
        assert_error(is_valid_var(name), '无效的参数名定义', lineno, line)
        defval = None
        if len(items) > 2:
            defval = items[2]

        if typ in TYPEMAPPING:
            typ = TYPEMAPPING[typ]
        p = cls(lineno, line, name, typ, defval, comment.strip())
        return p
    
    def __repr__(self):
        return '%s %s %s # %s\n' % (self.name, self.typ, self.defval, self.comment)

    def varcode(self):
        line = '%s %s // %s' % (self.name, self.typ, self.comment)
        return line

    def paramcode(self):
        defval = self.defval
        if self.typ == 'string':
            if defval is None:
                defval = '""'
            else:
                defval = '"%s"' % self.defval
        elif self.typ == 'bool':
            if defval is None:
                defval = false
        else:
            if defval is None:
                defval = 0

        line = 'param.%sVar(&%s, "%s", %s, "%s")' % (upperFirst(self.typ), self.name, self.name, defval, self.comment)
        return line

class CmdDef(object):
    def __init__(self, lineno, line, name, comment):
        self.lineno = lineno
        self.line = line
        self.name = name
        self.comment = comment
        self.cmdParams = []

    def __repr__(self):
        txt = '%s # %s\n' % (self.name, self.comment)
        for param in self.cmdParams:
            isinstance(param, CmdParam)
            txt += ' ' * TAB_WIDTH * 2 + str(param)
        return txt

    def addCmdParam(self, cmdParam):
        assert isinstance(cmdParam, CmdParam)
        self.cmdParams.append(cmdParam)

    def code(self):
        vars = ''
        params = '''
        param := flag.NewFlagSet("{name}", flag.ExitOnError)\n
        '''.format(name=self.name)

        for param in self.cmdParams:
            vars += '%s' % param.varcode()
            params += '%s' % param.paramcode()
        code = CMD_TPL.format(name=self.name, desc=self.comment, vars=vars, params=params)
        return code

    @classmethod
    def fromLine(cls, lineno, line):
        if '#' in line:
            itemsline, comment = line.split('#', 1)
            items = itemsline.split()
        else:
            comment = ''
            items = line.split()
        assert_error(len(items) >= 1, '无效的参数定义', lineno, line)
        name = items[0]
        assert_error(is_valid_var(name), '无效的参数名定义', lineno, line)
        p = cls(lineno, line, name, comment.strip())
        return p

class CmdModel(object):
    def __init__(self, root, cmds):
        self.root = root
        self.cmds = cmds

    def __repr__(self):
        txt = 'root cmd\n'
        for cmd in self.cmds:
            isinstance(cmd, CmdDef)
            txt += str(cmd)
        return txt

    def genCmdCode(self):
        for cmd in self.cmds:
            code = cmd.code()
            cmdPath = '%s/cmd/cmd_%s.go' % (MOD, cmd.name)
            if os.path.exists(cmdPath):
                print(Color.red('%s already exists! skip gen code' % cmdPath))
                continue
            writeCode(cmdPath, code)
        os.system('gofmt -w %s/cmd' % MOD)

    @classmethod
    def fromLines(cls, lines):
        lineno = 1
        root = lines[0].strip()
        assert_error(is_valid_var(root[0]), '开头必须是有效字母', lineno, root)
        items = root.split(' ')
        assert_error(len(items) >= 2, '无效的cmdmodel', lineno, root)
        assert_error(items[0] == 'root', '无效的cmdmodel', lineno, root)
        PREFIX = items[1]
        assert_error(PREFIX == 'cmd', '关键字必须是有效字母', lineno, root)

        cmds = []
        currentCmd = None

        for line in lines[1:]:
            lineno += 1

            if not line.strip():
                continue

            width = len(line) - len(line.lstrip())
            assert_error(width % TAB_WIDTH == 0, '宽带不是%s的倍数' % TAB_WIDTH, lineno, line)
            deep = int(width / TAB_WIDTH)
            assert_error(deep <= 2, '深度不能超过2', lineno, line)

            line = line.strip()
            if line.startswith('#'):
                continue
            # cmd
            if deep == 0:
                currentCmd = CmdDef.fromLine(lineno, line)
                cmds.append(currentCmd)
            # param
            elif deep == 1:
                assert currentCmd
                currentParam = CmdParam.fromLine(lineno, line)
                currentCmd.addCmdParam(currentParam)

        return cls(root, cmds)


if __name__ == "__main__":
    text = '''root cmd
hello # 支持4个类型, 字符串, 整数, 布尔, 浮点 -> name, desc
    name str # 姓名, 必须, 字符串 -> arg, desc
    age int 18 # 年纪, 可选, 默认18, 数字
    isboy bool false # 是否男孩, 默认false
    height float # 身高, 必须
createIndex
    '''
    cm = CmdModle.fromLines(text.splitlines())
    print(cm)
    cm.genCmdCode()




#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2019.07.20

class SwaggerSchema(object):
    def __init__(self, name, type, ref, description):
        self.name = name
        self.type = type # array / object
        self.ref = ref
        self.description = description

        self.required = [] # param_name in properties
        self.properties = {} # param_name : propertie
        self.items = None # type array's item schema

    def addRequiredParam(self, name):
        self.required.append(name)

    def addPropertieItem(self, name, schema):
        assert isinstance(schema, SwaggerSchema)
        self.properties[name] = schema

    def setArrayItem(self, schema):
        assert isinstance(schema, SwaggerSchema)
        self.type = 'array'
        self.items = schema

    def todict(self):
        d = {'type': self.type, 'description': self.description}
        if self.ref:
            d['$ref'] = self.ref
        elif self.items:
            d['items'] = self.items.todict()

        if len(self.required) > 0:
            d['required'] = self.required

        dp = {}
        for k, v in self.properties.items(): # py2 vs py3
            dp[k] = v.todict()
        if len(dp) > 0:
            d['properties'] = dp

        return d

    def clone(self):
        c = SwaggerSchema(self.name, self.type, self.ref, self.description)
        c.required = list(self.required)
        c.properties = dict(self.properties)
        c.items = self.items
        return c



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28
from __future__ import print_function
import os
import sys
import json
import zipfile

from .conf import *
from .swagger import *
from .helper import *
from .version import version
from .service import new_srv_code
from .cmdmodel import CmdModel

if M2C_VERSION == 'v4':
    from .apimodel_v4 import ApiModelV4 as ApiModel
    from .swagger import genSwaggerGwGoV4

if IS_PY3():
    from urllib.request import urlretrieve
else:
    from urllib import urlretrieve


def swaggerhtml(apimodelPath):
    from .swagger3 import genSwaggerContent
    yaml = genSwaggerContent(apimodelPath)
    return yaml

def goswagger3():
    docpath = os.path.join(GOPHER_DIR, 'ak/gw/swagger.yaml')
    godocpath = os.path.join(GOPHER_DIR, 'ak/gw/swagger.go')
    from .swagger3 import genSwaggerYaml, genSwaggerGwGo
    if M2C_VERSION == 'v4':
        docpath = SWAGGER_YAML_PATH
        godocpath = SWAGGER_PATH
        from .swagger3 import genSwaggerGwGoV4 as genSwaggerGwGo

    genSwaggerYaml(docpath)
    genSwaggerGwGo(godocpath)
    print(Color.green('Gen %s done.' % docpath))
    print(Color.green('Gen %s done.' % godocpath))
 
def goswagger():
    config = {
        "name": "myapp",
        "version": "0.0.1",
        "description": "myapp api doc",
        "title": "myapp api doc",
        "url": "https://myapp.com"
    }

    if os.path.exists('dev.yaml'):
        try:
            from yaml import load, Loader
            app = load(open('dev.yaml'), Loader)
            config['name'] = app['name']
            config['description'] = app['apiDoc']['desc']
            config['title'] = app['apiDoc']['title']
            config['url'] = app['apiDoc']['url']
        except:
            pass
    url = config['url']
    
    docpath = os.path.join(GOPHER_DIR, 'ak/gw/swagger.go')
    genSwaggerGwGo(docpath, url, config['title'], config['description'], config['version'])
    print(Color.green('Gen goswagger to %s done.' % docpath))
    
def goapidoc():
    lines = open(APIMODEL_PATH, 'rb').readlines()
    if IS_PY3():
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    apiModel = ApiModel.fromLines(lines)
    doc = apiModel.apiDoc()
    docpath = os.path.join(APIDOC_DIR, 'doc.go')
    writeCode(docpath, doc)
    print(Color.green('Gen api doc to %s done.' % docpath))

def apicode():
    lines = open(APIMODEL_PATH, 'rb').readlines()
    if IS_PY3():
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    apiModel = ApiModel.fromLines(lines)
    apiModel.genApiCode()
    update_m2c(version, M2C_VERSION)
    print(Color.green('Gen api code done.'))
    print(Color.red('不要修改包含<// Code generated by m2c. DO NOT EDIT.> 头的自动生成文件, 会被覆盖!!!'))

def apicodev2():
    lines = open(APIMODEL_PATH, 'rb').readlines()
    if IS_PY3():
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    apiModel = ApiModel.fromLines(lines)
    apiModel.genApiCodeV2()
    update_m2c(version, M2C_VERSION)
    print(Color.green('Gen api code done.'))
    print(Color.red('不要修改包含<// Code generated by m2c. DO NOT EDIT.> 头的自动生成文件, 会被覆盖!!!'))

def objcode():
    lines = open(OBJMODEL_PATH, 'rb').readlines()
    if IS_PY3(): # py2 vs py3
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    objModel = ObjModel.fromLines(lines)
    objModel.checkObjCode()
    if M2C_VERSION == 'v4':
        objModel.genObjCodeV4()
    else:
        objModel.genObjCode()
    update_m2c(version, M2C_VERSION)
    print(Color.green('Gen object code done.'))
    print(Color.red('不要修改包含<// Code generated by m2c. DO NOT EDIT.> 头的自动生成文件, 会被覆盖!!!'))

def pubm2czip(v='0.0.1'):
    name = 'm2c-%s.zip' % v
    output = os.path.join(BUILD_DIST, name)
    tar = 'git archive --format zip --output %s release/%s' % (output, v)
    assert os.system(tar) == 0, 'archive %s failed!' % output
    up = 's3cmd put %s s3://dat-pub/m2c/%s' % (output, name)
    assert os.system(up) == 0, 'run %s failed!' % up
    print(Color.green('pub m2c %s success!' % name))

def check_m2c():
    if is_m2c():
        return
    print(Color.red("It's not m2c project, init first!!!"))
    print(Color.green('Run: m2c -h for a help'))
    print(Color.green('Run: mkdir <your-project-name> to create an empty project dir'))
    print(Color.green('Run: cd <your-project-name> enter your project dir'))
    print(Color.green('Run: m2c init'))
    os._exit(-1)


def show_m2c():
    data = open('.m2c').read()
    print(json.dumps(json.loads(data), indent=2))

def init(git=''):
    if os.path.exists('.m2c'):
        print(Color.red("It's already m2c project"))
        show_m2c()
        os._exit(-1)
    
    if len(os.listdir('.')) != 0:
        print(Color.red('Current dir not empty, create your project dir!'))

    project_name = os.path.basename(os.getcwd())
    print(Color.green('Start init...'))
    print(Color.green('Project name: %s' % project_name))

    if git != '':
        print(Color.green('Clone from %s' % git))
        os.system('git clone %s .' % git)
    else:
        m2c_work_dir = '__m2c__'
        m2c_zip_path = os.path.join(m2c_work_dir, M2C_PROJECT_INIT_NAME)
        os.mkdir(m2c_work_dir)
        zipurl = M2C_PROJECT_INIT_URL
        if M2C_VERSION == 'v4':
            zipurl = M2C_PROJECT_INIT_URL_V4
        urlretrieve(zipurl, m2c_zip_path)
        print(Color.green('Download %s success!' % zipurl))
        zip_m2c = zipfile.ZipFile(m2c_zip_path, 'r')
        zip_m2c.extractall('.')
        zip_m2c.close()
        os.system('rm -rf %s' % m2c_work_dir)

    cfg = {
        'name': project_name,
        '__version__': version,
        '__cliversion__': M2C_VERSION,
        '__msg__': 'this is m2c init record file, keep this here!',
    }
    writeCode('.m2c', json.dumps(cfg, indent=2))
    sed_file(APP_DEV_CONFIG_FILE, 'name: "gop"\n', 'name: "%s"\n' % project_name) # update project name
    print(Color.green('m2c project %s init success!' % project_name))

def gen_service(name, force=False):
    srv_dir = os.path.join(SERVICE_DIR, name)
    srv_path = os.path.join(srv_dir, '%s.go' % name)
    srv_init_path = os.path.join(srv_dir, 'init.go')
    if os.path.exists(srv_init_path) and not force:
        print(Color.red('%s already exists' % srv_init_path))
        os._exit(-1)

    init_code, srv_code = new_srv_code(name)
    writeCode(srv_init_path, init_code)
    writeCode(srv_path, srv_code)
    os.system('gofmt -w %s/service/%s' % (MOD, name))
    print(Color.green('init service %s done' % name))


def cmdcode():
    lines = open(CMDMODEL_PATH, 'rb').readlines()
    if IS_PY3():
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    cmdModel = CmdModel.fromLines(lines)
    cmdModel.genCmdCode()
    print(Color.green('Gen cmd code done.'))



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28

from __future__ import print_function
import os
import sys
import json
import time

from .helper import IS_PY3, is_macos
from .conf import *
from .m2c import swaggerhtml
if M2C_VERSION == 'v3':
    from .m2c import goswagger3 as goswagger
elif M2C_VERSION == 'v4':
    from .m2c import goswagger3 as goswagger
else:
    from .m2c import goswagger

# amd64 only
os.environ.update({
    "CGO_ENABLED": "0",
    "GOARCH": "amd64"
})

# auto gen build details
date = time.strftime('%Y%m%d', time.localtime())
head = os.popen('git rev-parse HEAD').read().strip()
ldflags = " ".join([
    '-extldflags -static',
    '-X main.BuildDate=%s' %  date,
    '-X main.BuildGitHash=%s' % head
])

def _get_version():
    return head[:6]

def _get_name():
    return json.loads(open(M2C_PATH).read())['name']


def _get_logdir():
    name = _get_name()
    return os.path.join(LOG_HOME, name)


def _run(cmd):
    print('running: %s' % cmd)
    code = os.system(cmd)
    if code != 0:
        print(Color.red('run: %s failed' % cmd))
    else:
        print(Color.green('run: %s success' % cmd))


def version():
    print(_get_version())


def name():
    print(_get_name())


def build():
    os.chdir(GOPHER_DIR)
    app = os.path.join(BUILD_DIST, _get_name())
    cmds = [
        'mkdir -p %s' % BUILD_DIST,
        'go mod vendor',
        'go build -mod=vendor -ldflags "%s" -o %s .' % (ldflags, app),
    ]
    list(map(_run, cmds))
    print('Done')


def build_linux():
    os.environ.update({
        "GOOS": "linux",
    })
    os.chdir(GOPHER_DIR)
    app = os.path.join(BUILD_DIST, _get_name())
    cmds = [
        'mkdir -p %s' % BUILD_DIST,
        'go mod vendor',
        'go build -mod=vendor -ldflags "%s" -o %s .' % (ldflags, app),
    ]
    list(map(_run, cmds))
    print('Done')


def pack(env='dev', goos='linux'):
    os.environ.update({
        "GOOS": goos,
    })
    build()
    name = _get_name()
    version = _get_version()
    dist = os.path.join(BUILD_DIST, '%s-%s.tar.gz' % (name, version))
    cfg = os.path.join(GOPHER_DIR, 'config', '%s.yaml' % env)
    os.system('cp %s %s' % (cfg, BUILD_DIST))
    files = " ".join([name, '%s.yaml' % env])

    os.chdir(BUILD_DIST)
    cmds = [
        'tar -zcf %s %s' % (dist, files)
    ]
    list(map(_run, cmds))

    print('name: %s' % name)
    print('version: %s' % version)
    print('pack: %s' %  dist)

def packsrc():
    name = _get_name()
    version = _get_version()
    dist = os.path.join('dist', '%s-%s-src.tar.gz' % (name, version))

    cmds = [
        'tar -zcf %s %s' % (dist, 'gopher')
    ]
    map(_run, cmds)

    print('name: %s' % name)
    print('version: %s' % version)
    print('pack: %s' %  dist)


def clean():
    os.system('rm -rf dist/*')


def unittest():
    os.chdir(GOPHER_DIR)
    cmds = [
        'cd test/ && docker-compose run --rm unittest'
    ]
    map(_run, cmds)
    print('Done')
    

def supervisor_conf(apphome='/data/app', env='dev'):
    conf = '''
[program:{APPNAME}]
directory={APPHOME}
command={APPBIN_PATH} -c {APPCONF_PATH}
autostart=true
autorestart=true
stdout_logfile={STDOUT_LOG_PATH}
stdout_logfile_maxbytes=50M
stderr_logfile={STDERR_LOG_PATH}
stderr_logfile_maxbytes=50M'''
    appname = _get_name()
    apphome = os.path.join(apphome, appname)
    appbin_path = os.path.join(apphome, appname)
    appconf_path = os.path.join(apphome, '%s.yaml' % env)
    logdir = _get_logdir()
    if not os.path.isabs(logdir):
        if logdir.startswith('./'):
            logdir = logdir.lstrip('./')
        logdir = os.path.join(apphome, logdir)
    stdout_log_path = os.path.join(logdir, '%s_stdout.log' % appname)
    stderr_log_path = os.path.join(logdir, '%s_stderr.log' % appname)
    print(conf.format(**{
        "APPNAME": appname, 
        "APPHOME": apphome, 
        "APPBIN_PATH": appbin_path, 
        "APPCONF_PATH": appconf_path,
        "STDOUT_LOG_PATH": stdout_log_path,
        "STDERR_LOG_PATH": stderr_log_path,
    }))


def run(*args):
    goswagger()
    vendor()
    from .watchgo import WatchGo
    os.chdir(GOPHER_DIR)
    command = ['go', 'run', '.']
    if args:
        command.extend(args)
    WatchGo(command, GOPHER_DIR).start_watch()

def debug():
    os.chdir(GOPHER_DIR)
    os.system('dlv debug .')

def vendor():
    os.chdir(GOPHER_DIR)
    os.system('go mod vendor')

def swaggerCodegen(lang, outputDir, pkg='', swaggerYamlPath=SWAGGER_YAML_PATH):
    if os.system('which swagger-codegen') != 0:
        print(Color.red('swagger-codgen 不存在, 安装参考: https://github.com/swagger-api/swagger-codegen'))
        if is_macos():
            print(Color.green('当前为macos 自动执行安装中...'))
            _run('brew install swagger-codegen')
        return
    if pkg == '':
        _run('swagger-codegen generate -i %s -l %s -o %s' % (swaggerYamlPath, lang, outputDir))
    else:
        _run('swagger-codegen generate -i %s -l %s -o %s -DpackageName=%s' % (swaggerYamlPath, lang, outputDir, pkg))

def swaggerHtmlDoc(apimodelPath, pkg=''):
    html = swaggerhtml(apimodelPath)
    savedir = os.path.dirname(apimodelPath)
    savepath = os.path.join(savedir, 'swagger.yaml')
    print(savedir, savepath)
    fp = open(savepath, 'wb')
    header = '# Code generated by m2c. DO NOT EDIT.\n'
    if IS_PY3():
        fp.write(header.encode('utf-8'))
        fp.write(html.encode('utf-8'))
    else:
        fp.write(header)
        fp.write(html)
    fp.close()

    swaggerCodegen('html2', savedir, pkg=pkg, swaggerYamlPath=savepath)



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2019.07.20

import os
import json
from .apimodel import ApiModel, ApiDef
from .objmodel import ObjModel
from .schema import SwaggerSchema
from .conf import Color, MOD, M2C_VERSION
from .helper import IS_PY3

if M2C_VERSION == 'v4':
    from .apimodel_v4 import ApiModelV4 as ApiModel
# OpenAPI Swargger https://swagger.io/specification

SWAGGER_INDEX = '''
<!-- HTML for static distribution bundle build -->
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8">
    <title>Swagger UI</title>
    <link rel="stylesheet" type="text/css" href="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.22.0/swagger-ui.css" >
    <style>
      html
      {
        box-sizing: border-box;
        overflow: -moz-scrollbars-vertical;
        overflow-y: scroll;
      }

      *,
      *:before,
      *:after
      {
        box-sizing: inherit;
      }

      body
      {
        margin:0;
        background: #fafafa;
      }
    </style>
  </head>

  <body>
    <div id="swagger-ui"></div>

    <script src="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.22.0/swagger-ui-bundle.js"> </script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.22.0/swagger-ui-standalone-preset.js"> </script>
    <script>
    window.onload = function() {
      // Begin Swagger UI call region
      let hash = window.location.hash
      let url = window.location.href
      url = url.replace(hash, '') + '.yaml'
      const ui = SwaggerUIBundle({
        url: url,
        dom_id: '#swagger-ui',
        deepLinking: true,
        presets: [
          SwaggerUIBundle.presets.apis,
          SwaggerUIStandalonePreset
        ],
        plugins: [
          SwaggerUIBundle.plugins.DownloadUrl
        ],
        layout: "StandaloneLayout"
      })
      // End Swagger UI call region

      window.ui = ui
    }
  </script>
  </body>
</html>
'''

class SwaggerInformation(object):
    def __init__(self, title, description, version):
        self.title = title
        self.description = description
        self.version = version

    def todict(self):
        return {'title': self.title, 'description': self.description, 'version': self.version}

class SwaggerTag(object):
    def __init__(self, name, description):
        self.name = name
        self.description = description

    def todict(self):
        return {'name': self.name, 'description': self.description}

class SwaggerServer(object):
    def __init__(self, url, description):
        self.url = url
        self.description = description

    def todict(self):
        return {'url': self.url, 'description': self.description}

class SwaggerPathItem(object):
    def __init__(self, path, description, postop=None, getop=None):
        self.path = path
        self.description = description
        self.postop = postop # post operation
        self.getop = getop # get operation

    def setGetOperation(self, op):
        assert isinstance(op, SwaggerOperation)
        self.getop = op

    def setPostOperation(self, op):
        assert isinstance(op, SwaggerOperation)
        self.postop = op

    def todict(self):
        d = {'description': self.description}
        if hasattr(self, 'getop') and self.getop:
            d['get'] = self.getop.todict()
        if hasattr(self, 'postop') and self.postop:
            d['post'] = self.postop.todict()
        return d

class SwaggerOperation(object):
    def __init__(self, summary, description, operationID):
        self.summary = summary # short desc
        self.description = description
        self.operationID = operationID # apiName
        self.tags = [] # apiGroup
        self.parameters = [] # get request in query, ref schema
        self.requestBody = None # post request, ref schema
        self.responses = None
        self.security = [] # operation security

    def addTag(self, tag):
        self.tags.append(tag)

    def addQueryParam(self, param):
        assert isinstance(param, SwaggerParameter)
        self.parameters.append(param)

    def setRequestBodyRef(self, contentType, ref):
        self.requestBody = {
            'content': {
                contentType: {
                    'schema': {'$ref': ref}
                }
            }
        }

    def setResponseRef(self, statusCode, description, contentType, ref):
        self.responses = {
            statusCode: {
                'description': description,
                'content': {
                    contentType: {'schema': {'$ref': ref}}
                }
            }
        }

    def todict(self):
        d = {'summary': self.summary, 'description': self.description, 'operationId': self.operationID}
        if len(self.tags) > 0:
            d['tags'] = self.tags
        if len(self.parameters) > 0:
            d['parameters'] = list(map(lambda x: x.todict(), self.parameters))
        if self.requestBody:
            d['requestBody'] = self.requestBody
        if self.responses:
            d['responses'] = self.responses
        if self.security:
            d['security'] = self.security
        return d

class SwaggerParameter(object):
    def __init__(self, location, name, description, schema, required=True):
        self.location = location
        self.name = name # param_name
        self.description = description
        self.required = required
        self.schema = schema

    def todict(self):
        d = {'in': self.location, 'name': self.name, 'description': self.description, 'required': self.required}
        d['schema'] = self.schema.todict()
        return d

class Swagger(object):
    def __init__(self, title, description, version):
        self.openapi = '3.0.0'
        self.info = SwaggerInformation(title, description, version)
        self.servers = []
        self.paths = {} # apis {"/login", SwaggerItem}
        self.tags = [] # groups
        self.definitions = {} # models, properties_name: properties
        self.security = [] # global security object
        self.securitySchemes = {}

    def addServer(self, url, description):
        self.servers.append(SwaggerServer(url, description))

    def addTag(self, name, description):
        self.tags.append(SwaggerTag(name, description))

    def addPath(self, name, sgItem):
        assert isinstance(sgItem, SwaggerPathItem)
        assert name not in self.paths
        self.paths[name] = sgItem

    def addDefinition(self, name, schema):
        assert isinstance(schema, SwaggerSchema)
        assert name not in self.definitions, Color.red('definition "%s" conflict.' % name)
        self.definitions[name] = schema

    def addSecuritySchemes(self, name, obj):
        if name in self.securitySchemes:
            return

        self.securitySchemes[name] = obj

    def getSecuritySchemes(self, auth, authmap):
        dstauth = authmap.get(auth, '')
        if dstauth == 'apikey':
            return self.getApiKeyAuth('apikey', 'header')
        if dstauth.startswith('apikey_'):
            name = dstauth.replace('apikey_', '')
            return self.getApiKeyAuth(name, 'header')
        elif dstauth == 'jwt':
            return self.getJWTAuth()
        elif dstauth == 'basic':
            return self.getBasicAuth()
        if dstauth.startswith('cookie_'):
            name = dstauth.replace('cookie_', '')
            return self.getApiKeyAuth(name, 'cookie')
    
    def getApiKeyAuth(self, name, location):
        return {
            'type': 'apiKey',
            'name': name,
            'in': location,
        }

    def getBasicAuth(self):
        return {
            'type': 'http',
            'scheme': 'basic',
        }

    def getJWTAuth(self):
        return {
            'type': 'http',
            'scheme': 'bearer',
            'bearerFormat': 'JWT',
        }

    def todict(self):
        d = {'openapi': self.openapi}
        dps = {}
        for k, v in self.paths.items():
            dps[k] = v.todict()
        d['paths'] = dps
        dds = {}
        for k, v in self.definitions.items():
            if isinstance(v, dict):
                dds[k] = v
            else:
                dds[k] = v.todict()
        d['components'] = {'schemas': dds, 'securitySchemes': self.securitySchemes}
        d['tags'] = list(map(lambda x: x.todict(), self.tags))
        d['servers'] = list(map(lambda x: x.todict(), self.servers)) # py2 vs py3
        if self.info:
            d['info'] = self.info.todict()
        return d

def genSwaggerContent(apimodelPath = 'apimodel.txt'):
    lines = open(apimodelPath, 'rb').readlines()
    if IS_PY3(): # py2 vs py3
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    apiModel = ApiModel.fromLines(lines)

    jsonRespIn = apiModel.resp['in']
    respContainerModel = SwaggerSchema('respContainer', 'object', '', '统一JSON返回结构体')
    for param in apiModel.resp['fields']:
        respContainerModel.addRequiredParam(param)
        if param == apiModel.resp['in']:
            respContainerModel.addPropertieItem(param, SwaggerSchema(param, 'object', '', ''))
        else:
            respContainerModel.addPropertieItem(param, SwaggerSchema(param, 'string', '', ''))

    # sg
    sg = Swagger(apiModel.title, apiModel.description, apiModel.srvVersion)
    sg.addServer(apiModel.apiServer, '')

    for group in apiModel.groups:
        sg.addTag(group.group, group.comment)

        for api in group.apis:
            path = SwaggerPathItem(api.apiName(), api.comment)
            op = SwaggerOperation(api.comment, api.comment, api.apiFullName())

            # 设置auth
            ssc = sg.getSecuritySchemes(api.auth, apiModel.authmap)
            if ssc:
                sg.addSecuritySchemes(api.auth, ssc)
                op.security.append({api.auth: []})

            reqContentType = 'application/json'
            respContentType = 'application/json'
            if api.proto == 'raw':
                reqContentType = 'application/octet-stream'
                respContentType = 'application/octet-stream'
            elif api.proto == 'reqRaw':
                reqContentType = 'application/octet-stream'
            elif api.proto == 'respRaw':
                respContentType = 'application/octet-stream'

            op.addTag(api.group)

            # api req model
            reqModel = SwaggerSchema(api.reqModelName(), 'object', '', api.comment)
            for param in api.apiParams:
                if param.required:
                    reqModel.addRequiredParam(param.name)
                reqModel.addPropertieItem(param.name, param.schema())

                if api.method.lower() == 'get':
                    qpm = SwaggerParameter('query', param.name, param.comment, param.schema(), param.required)
                    op.addQueryParam(qpm)

            if api.method.lower() != 'get':
                sg.addDefinition(api.reqModelName(), reqModel)

            # api resp model
            respModel = SwaggerSchema(api.respModelName(), 'object', '', api.comment)
            for resp in api.apiResps:
                if resp.required:
                    respModel.addRequiredParam(resp.name)
                respModel.addPropertieItem(resp.name, resp.schema())

            if respContentType == 'application/json':
                jm = respContainerModel.clone()
                if len(respModel.properties) == 0:
                    respModel.type = 'string' # 兼容swagger-codegen 空对象对生成sdk序列化有影响
                jm.addPropertieItem(jsonRespIn, respModel)
                sg.addDefinition(api.respModelName(), jm)
            else:
                sg.addDefinition(api.respModelName(), respModel)

            # request param
            if api.method.lower() != 'get':
                op.setRequestBodyRef(reqContentType, '#/components/schemas/%s' % api.reqModelName())

            # response
            op.setResponseRef(200, api.comment, respContentType, '#/components/schemas/%s' % api.respModelName())

            if api.method.lower() == 'get':
                path.setGetOperation(op)
            else:
                path.setPostOperation(op)

            apiName = '/api/%s/%s/%s' % (apiModel.version, api.group, api.apiRouterName())
            sg.addPath(apiName, path)

    d = sg.todict()
    from yaml import safe_dump
    return safe_dump(d, indent=4)

def genSwaggerYaml(savePath):
    content = genSwaggerContent()
    fp = open(savePath, 'wb')
    header = '# Code generated by m2c. DO NOT EDIT.\n'
    if IS_PY3():
        fp.write(header.encode('utf-8'))
        fp.write(content.encode('utf-8'))
    else:
        fp.write(header)
        fp.write(content)
    fp.close()

def genSwaggerGwGo(savePath):
    content = genSwaggerContent()
    fp = open(savePath, 'wb')
    code = '''
// Code generated by m2c. DO NOT EDIT.
package gw

const SWAGGER_INDEX = `
%s
`

const SWAGGER_CONTENT = `
%s
`
    ''' % (SWAGGER_INDEX, content)
    if IS_PY3():
        fp.write(code.encode('utf-8'))
    else:
        fp.write(code)
    fp.close()
    os.system('gofmt -w %s/' % MOD)

def genSwaggerGwGoV4(savePath):
    content = genSwaggerContent()
    fp = open(savePath, 'wb')
    code = '''
// Code generated by m2c. DO NOT EDIT.
package router

const SWAGGER_INDEX = `
%s
`

const SWAGGER_CONTENT = `
%s
`
    ''' % (SWAGGER_INDEX, content)
    if IS_PY3():
        fp.write(code.encode('utf-8'))
    else:
        fp.write(code)
    fp.close()
    os.system('gofmt -w %s/' % MOD)



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28
 
import os


M2C_VERSION = 'v2'
MOD = 'gopher'
PROJECT_DIR = os.getcwd()
GOPHER_DIR = os.path.join(PROJECT_DIR, MOD)
APP_DIR = os.path.join(GOPHER_DIR, 'app')
APP_DEV_CONFIG_FILE = os.path.join(GOPHER_DIR, 'config/dev.yaml')
SERVICE_DIR = os.path.join(GOPHER_DIR, 'service')
APIDOC_DIR = os.path.join(PROJECT_DIR, 'doc/api')
M2C_PATH = os.path.join(PROJECT_DIR, '.m2c')
BUILD_DIST = os.path.join(PROJECT_DIR, 'dist')
SWAGGER_PATH = os.path.join(GOPHER_DIR, "app/router/swagger.go")
SWAGGER_YAML_PATH = os.path.join(GOPHER_DIR, "app/router/swagger.yaml")
GOMOD_PATH = os.path.join(GOPHER_DIR, 'go.mod')
APIMODEL_PATH = os.path.join(PROJECT_DIR, 'apimodel.txt')
OBJMODEL_PATH = os.path.join(PROJECT_DIR, 'objmodel.txt')
CMDMODEL_PATH = os.path.join(PROJECT_DIR, 'cmdmodel.txt')
M2C_PROJECT_INIT_URL = 'https://dat-pub.oss-cn-hangzhou.aliyuncs.com/m2c/m2c-0.0.3.zip'
M2C_PROJECT_INIT_URL_V4 = 'https://dat-pub.oss-cn-hangzhou.aliyuncs.com/m2c/m2c-0.1.3.zip'
M2C_PROJECT_INIT_NAME = os.path.basename(M2C_PROJECT_INIT_URL)
LOG_HOME = '/data/log'

TYPEMAPPING = {
    'str': 'string',
    '[]str': '[]string',
    'string': 'string',
    '[]string': '[]string',
    'int': 'int64',
    'int32': 'int32',
    'int64': 'int64',
    'long': 'int64',
    'float': 'float64',
    'float64': 'float64',
    'float32': 'float32',
    'double': 'float64',
    '[]int': '[]int64',
    'int32': 'int32',
    '[]int32': '[]int32',
    'int64': 'int64',
    '[]int64': '[]int64',
    'long': 'int64',
    '[]long': '[]int64',
    'float': 'float64',
    '[]float': '[]float64',
    'float64': 'float64',
    '[]float64': '[]float64',
    'float32': 'float32',
    '[]float32': '[]float32',
    'double': 'float64',
    '[]double': '[]float64',
    'time': 'time.Time',
    'bool': 'bool',
    '[]bool': '[]bool',
    'any': 'interface{}',
    'dict': 'map[string]interface{}',
    'json': 'map[string]interface{}',
    'JsonNode': 'map[string]interface{}',
}

APIDOC_TYPE = {
    'int64': 'Number',
    'int32': 'Number',
    'int': 'Number',
    'float': 'Number',
    'float32': 'Number',
    'float64': 'Number',
    '[]float': 'Number[]',
    '[]float32': 'Number[]',
    '[]float64': 'Number[]',
    '[]int64': 'Number[]',
    '[]int32': 'Number[]',
    '[]int': 'Number[]',
    'str': 'String',
    '[]str': 'String[]',
    'string': 'String',
    '[]string': 'String[]',
    'interface{}': 'Object',
    '[]interface{}': 'Object[]',
    'bool': 'Boolean',
    '[]bool': 'Boolean[]',
    'object': 'Object',
    '[]object': 'Object[]',
    'number': 'Number',
    '[]number': '[]Number',
    'number[]': 'Number[]', # old
    'object[]': 'Object[]', # old
}

OPENAPIDOC_TYPE = {
    'int64': 'integer',
    'int32': 'integer',
    'int': 'integer',
    'float': 'number',
    'float32': 'number',
    'float64': 'number',
    'str': 'string',
    'string': 'string',
    'interface{}': 'object',
    'bool': 'boolean',
    'object': 'object',
    'number': 'number',
}

AUTH_TYPE = ['userAuth', 'adminAuth', None]
PROTO_TYPE = ['json', 'raw', 'reqRaw', 'respRaw', None]

TAB_WIDTH = 4
API_MAX_DEEP_NUM = 2


OBJ_ALLOWED_PREFIX = ["&", "@", "^"]
OBJ_REF_TYPE_MANAGE = "manage"
OBJ_REF_TYPE_REF = "refer"
OBJ_CONTAINER_TYPE_BASIC = "basic"
OBJ_CONTAINER_TYPE_SINGLE = "single"
OBJ_CONTAINER_TYPE_LIST = "list"
OBJ_CONTAINER_TYPE_TREE = "tree" # TODO
OBJ_MAX_DEEP_NUM = 3

class Color:
    _GREEN = "\033[92m"
    _RED = "\033[91m"
    _BOLD = '\033[1m'
    _ENDC = '\033[0m'

    @staticmethod
    def red(txt):
        return "%s%s%s%s" % (Color._BOLD, Color._RED, txt, Color._ENDC)

    @staticmethod
    def green(txt):
        return "%s%s%s%s" % (Color._BOLD, Color._GREEN, txt, Color._ENDC)

def set_version(ver):
    global M2C_VERSION
    M2C_VERSION = ver



from __future__ import print_function
import os
import sys
from .conf import *
from .helper import *

if sys.argv[-1] == '-v3' or sys.argv[-1] == '-v4':
    set_version(sys.argv[-1].strip('-'))
    sys.argv = sys.argv[:-1]
else:
    set_version(m2c_cliversion())

print('m2c_version:', m2c_version())
print('m2c_cliversion:', M2C_VERSION)
print('py3_version:', IS_PY3())

if os.path.exists('.env'):
    for line in open('.env'):
        items = line.split('=')
        if len(items) < 2:  # hander '=' in value(query string)
            continue
        print('load env:', items[0], '='.join(items[1:]))
        os.environ[items[0].strip()] = '='.join(items[1:]).strip()



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28

OBJ_FIXED_CODE_TPL = '''
func Get$BIGOBJNAME$Table() *mgo.Collection {
	return $OBJNAME$Table
}

func Get(id int64) ($OBJNAME$ *$BIGOBJNAME$, err error) {
	$OBJNAME$ = &$BIGOBJNAME${}
	err = $OBJNAME$Table.FindId(id).One($OBJNAME$)
	if err != nil && err.Error() == "not found" {
		$OBJNAME$ = nil
		err = nil
	}
	return
}

func GetResultTo(id int64, out interface{}) (err error) {
	err = $OBJNAME$Table.FindId(id).One(out)
	if err != nil && err.Error() == "not found" {
		out = nil
		err = nil
	}
	return
}

func MustGet(id int64, outerr error) ($OBJNAME$ *$BIGOBJNAME$) {
	$OBJNAME$ = &$BIGOBJNAME${}
	err := $OBJNAME$Table.FindId(id).One($OBJNAME$)
	if err != nil && err.Error() == "not found" {
		$OBJNAME$ = nil
		err = nil
	}

	if err != nil {
		panic(err)
	} 

	if $OBJNAME$ == nil {
		panic(outerr)
	}
	return
}

func MustGetResultTo(id int64, out interface{}, outerr error)  {
	err := $OBJNAME$Table.FindId(id).One(out)
	if err != nil && err.Error() == "not found" {
		out = nil
		err = nil
	}

	if err != nil {
		panic(err)
	} 

	if out == nil {
		panic(outerr)
	}
	return
}

func Save($OBJNAME$ *$BIGOBJNAME$) (err error) {
	if $OBJNAME$.Id == 0 {
		$OBJNAME$.Id = id.Gen("$OBJNAME$")
	}
	now := time.Now().Unix()
	if $OBJNAME$.CTime == 0 {
		$OBJNAME$.CTime = now
	}
	$OBJNAME$.MTime = now
	_, err = $OBJNAME$Table.UpsertId($OBJNAME$.Id, $OBJNAME$)
	return
}

func Update(id int64, updater map[string]interface{}) ($OBJNAME$ *$BIGOBJNAME$, err error) {
	now := time.Now().Unix()
	updater["mtime"] = now
	change := mgo.Change{
		Update:    bson.M{"$set": updater},
		Upsert:    true,
		ReturnNew: true,
	}
	$OBJNAME$ = &$BIGOBJNAME${}
	_, err = $OBJNAME$Table.Find(bson.M{"_id": id}).Apply(change, $OBJNAME$)
	if err != nil {
		$OBJNAME$ = nil
	}
	return
}

func Delete(id int64) (err error) {
	err = $OBJNAME$Table.RemoveId(id)
	if err != nil && err.Error() == "not found" {
		err = nil
	}
	return
}

type $BIGOBJNAME$Query struct {
	pageIndex int
	pageSize  int
	filter    bson.M
	or        []bson.M
	sort      []string
}

func Query() *$BIGOBJNAME$Query {
	return &$BIGOBJNAME$Query{
		filter: bson.M{},
		or:     []bson.M{},
	}
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) GetQuery() bson.M {
	if len($OBJNAME$Query.or) > 0 {
		$OBJNAME$Query.filter["$or"] = $OBJNAME$Query.or
	}
	return $OBJNAME$Query.filter
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) WithPage(pageIndex int64, pageSize int64) *$BIGOBJNAME$Query {
	$OBJNAME$Query.pageIndex = int(pageIndex-1) * int(pageSize)
	$OBJNAME$Query.pageSize = int(pageSize)
	return $OBJNAME$Query
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) WithFilter(filter map[string]interface{}) *$BIGOBJNAME$Query {
	if len($OBJNAME$Query.filter) == 0 {
		$OBJNAME$Query.filter = filter
	} else {
		for k, v := range filter {
			$OBJNAME$Query.filter[k] = v
		}
	}
	return $OBJNAME$Query
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) WithIn(key string, values interface{}) *$BIGOBJNAME$Query {
	$OBJNAME$Query.filter[key] = bson.M{"$in": values}
	return $OBJNAME$Query
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) WithNotIn(key string, values interface{}) *$BIGOBJNAME$Query {
	$OBJNAME$Query.filter[key] = bson.M{"$nin": values}
	return $OBJNAME$Query
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) WithSort(sort []string) *$BIGOBJNAME$Query {
	$OBJNAME$Query.sort = sort
	return $OBJNAME$Query
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) WithTextSearch(text string) *$BIGOBJNAME$Query {
	$OBJNAME$Query.filter["$text"] = bson.M{"$search": text}
	return $OBJNAME$Query
}

// https://docs.mongodb.com/manual/reference/operator/query/or/#or-and-text-queries
func ($OBJNAME$Query *$BIGOBJNAME$Query) OrTextSearch(text string) *$BIGOBJNAME$Query {
	$OBJNAME$Query.or = append($OBJNAME$Query.or, bson.M{"$text": bson.M{"$search": text}})
	return $OBJNAME$Query
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) Count() (n int64, err error) {
	_n, err := $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Count()
	n = int64(_n)
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) Update(updater map[string]interface{}) (updated int64, err error) {
	now := time.Now().Unix()
	updater["mtime"] = now
	change := bson.M{"$set": updater}
	info, err := $OBJNAME$Table.UpdateAll($OBJNAME$Query.GetQuery(), change)
	if err != nil {
		return
	}
	updated = int64(info.Updated)
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) One() ($OBJNAME$ *$BIGOBJNAME$, err error) {
	$OBJNAME$ = &$BIGOBJNAME${}
	err = $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Sort($OBJNAME$Query.sort...).One($OBJNAME$)
	if err != nil && err.Error() == "not found" {
		$OBJNAME$ = nil
		err = nil
	}
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) OneResultTo(out interface{}) (err error) {
	err = $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Sort($OBJNAME$Query.sort...).One(out)
	if err != nil && err.Error() == "not found" {
		out = nil
		err = nil
	}
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) MustOne(outerr error) ($OBJNAME$ *$BIGOBJNAME$) {
	$OBJNAME$ = &$BIGOBJNAME${}
	err := $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Sort($OBJNAME$Query.sort...).One($OBJNAME$)
	if err != nil && err.Error() == "not found" {
		$OBJNAME$ = nil
		err = nil
	}

	if err != nil {
		panic(err)
	}  

	if $OBJNAME$ == nil {
		panic(outerr)
	}
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) MustOneResultTo(out interface{}, outerr error) {
	err := $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Sort($OBJNAME$Query.sort...).One(out)
	if err != nil && err.Error() == "not found" {
		out = nil
		err = nil
	}

	if err != nil {
		panic(err)
	} 

	if out == nil {
		panic(outerr)
	}
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) All() ($OBJNAME$ []*$BIGOBJNAME$, err error) {
	$OBJNAME$ = []*$BIGOBJNAME${}
	err = $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Skip($OBJNAME$Query.pageIndex).Limit($OBJNAME$Query.pageSize).Sort($OBJNAME$Query.sort...).All(&$OBJNAME$)
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) AllResultTo(out interface{}) (err error) {
	err = $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Skip($OBJNAME$Query.pageIndex).Limit($OBJNAME$Query.pageSize).Sort($OBJNAME$Query.sort...).All(out)
	return
}

func ($OBJNAME$Query *$BIGOBJNAME$Query) Scan() (ch chan *$BIGOBJNAME$, err chan error) {
	iter := $OBJNAME$Table.Find($OBJNAME$Query.GetQuery()).Iter()
	ch = make(chan *$BIGOBJNAME$)
	err = make(chan error)
	go func() {
		defer iter.Close()
		defer close(ch)
		defer close(err)
		for {
			p := &$BIGOBJNAME${}
			if ok := iter.Next(p); ok {
				if p.CTime != 0 { // empty object
					ch <- p
				}
			}
			if iter.Err() != nil {
				err <- iter.Err()
				return
			}
			if iter.Done() {
				return
			}
		}
	}()
	return
}

// save $OBJNAME$
func ($OBJNAME$ *$BIGOBJNAME$) Save() (err error) {
	if $OBJNAME$.Id == 0 {
		$OBJNAME$.Id = id.Gen("$OBJNAME$")
	}
	now := time.Now().Unix()
	if $OBJNAME$.CTime == 0 {
		$OBJNAME$.CTime = now
	}
	$OBJNAME$.MTime = now
	_, err = $OBJNAME$Table.UpsertId($OBJNAME$.Id, $OBJNAME$)
	return
}
'''
OBJ_REF_HELPER_CODE_TPL = """
// single refer relation
func ({objName} *{ObjName}) Ref{RefName}({refName}Id int64) *{ObjName} {{
    {objName}.Ref{RefName}Id = {refName}Id
    return {objName}
}}

func ({objName} *{ObjName}) Get{RefName}() ({refName} *{refName}Object.{RefName}, err error) {{
    return {refName}Object.Get({objName}.Ref{RefName}Id)
}}
"""
OBJ_MGR_HELPER_CODE_TPL = '''
// manager relation, TODO transactions + upsert
func ({objName} *{ObjName}) Add{MgrShortName}({mgrShortName} *{mgrLongName}Object.{MgrLongName}) (err error) {{
    {mgrShortName}.Ref{ObjName}Id = {objName}.Id
    err = {mgrLongName}Object.Save({mgrShortName})
    if err != nil {{
        return
    }}
    {objName}.Mgr{MgrShortName}Id = {mgrShortName}.Id
    updater := map[string]interface{{}}{{
        "{mgrShortName}Id": {objName}.Mgr{MgrShortName}Id,
    }}
    {objName}, err = Update({objName}.Id, updater)
    return
}}

func ({objName} *{ObjName}) Get{MgrShortName}() ({mgrShortName} *{mgrLongName}Object.{MgrLongName}, err error) {{
    return {mgrLongName}Object.Get({objName}.Mgr{MgrShortName}Id)
}}
'''

OBJ_MGR_ARRAY_HELPER_CODE_TPL = '''
// array manager
func ({objName} *{ObjName}) Add{MgrShortName}({mgrShortName} *{mgrLongName}Object.{MgrLongName}) (err error) {{
	{mgrShortName}.Ref{ObjName}Id = {objName}.Id
	err = {mgrShortName}.Save()
	return
}}

func ({objName} *{ObjName}) Get{MgrShortName}ById({mgrShortName}Id int64) ({mgrShortName} *{mgrLongName}Object.{MgrLongName}, err error) {{
	return {mgrLongName}Object.Get({mgrShortName}Id)
}}

func ({objName} *{ObjName}) Query{MgrShortName}() (query{MgrShortName} *{mgrLongName}Object.{MgrLongName}Query) {{
	return {mgrLongName}Object.Query().WithFilter(map[string]interface{{}}{{
		"{objName}Id": {objName}.Id,
	}})
}}
'''
OBJ_MGR_REF_HELPER_CODE_TPL = '''
func ({objName} *{ObjName}) GetRef{ParentName}({parentName} interface{{}}) (err error) {{
	err = db.C("{parentName}").FindId({objName}.Ref{ParentName}Id).One({parentName})
	if err != nil && err.Error() == "not found" {{
		err = nil
	}}
	return
}}
'''
OBJ_FILTER_HELPER_CODE_TPL = '''
func ({objName}Query *{ObjName}Query) With{FieldName}Equal({fieldName} {filedType}) *{ObjName}Query {{
	{objName}Query.filter["{fieldName}"] = {fieldName}
	return {objName}Query
}}

func ({objName}Query *{ObjName}Query) With{FieldName}NotEqual({fieldName} {filedType}) *{ObjName}Query {{
	{objName}Query.filter["{fieldName}"] =  bson.M{{"$ne": {fieldName}}}
	return {objName}Query
}}

func ({objName}Query *{ObjName}Query) Or{FieldName}Equal({fieldName} {filedType}) *{ObjName}Query {{
	{objName}Query.or = append({objName}Query.or, bson.M{{"{fieldName}": {fieldName}}})
	return {objName}Query
}}
'''

OBJ_FILTER_STRING_LIKE_CODE_TPL = '''
func ({objName}Query *{ObjName}Query) Or{FieldName}Like({fieldName} string) *{ObjName}Query {{
	{objName}Query.or = append({objName}Query.or, bson.M{{"{fieldName}": bson.M{{"$regex": {fieldName}}}}})
	return {objName}Query
}}
func ({objName}Query *{ObjName}Query) With{FieldName}Like({fieldName} string) *{ObjName}Query {{
	{objName}Query.filter["{fieldName}"] = bson.M{{"$regex": {fieldName}}}
	return {objName}Query
}}
'''



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28
from __future__ import print_function
import os

from .conf import *
from .helper import *
from .objtpl import *

SERVICE_TPL = '''
package {PKG_NAME}

import (
	"sync"
)

type {SRV_NAME}Service struct {{
}}

var (
	_{SRV_NAME_LOW}Service *{SRV_NAME}Service
	_once        sync.Once
)

func Init() {{
	_once.Do(func() {{
		_{SRV_NAME_LOW}Service = &{SRV_NAME}Service{{
		}}
	}})
}}

func GetInstance() *{SRV_NAME}Service {{
	if _{SRV_NAME_LOW}Service == nil {{
		panic("{SRV_NAME_LOW} service not init!!!")
	}}
	return _{SRV_NAME_LOW}Service
}}'''

SERVICE_EMPTY = '''
package {PKG_NAME}
'''

# srv 服务名, 小驼峰格式, 如user, eventSystem
def new_srv_code(srv):
    pkg = srv
    srv_name = upperFirst(srv)
    return SERVICE_TPL.format(PKG_NAME=pkg, SRV_NAME=srv_name, SRV_NAME_LOW=srv), SERVICE_EMPTY.format(PKG_NAME=pkg)



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28
import os

from .conf import *
from .schema import SwaggerSchema
from .helper import *

# Api model to code
'''
root apiPrefix
apiGroup # comment
    apiName [method=get] [auth='<your any auth name>'] [proto=[json, raw, reqRaw, respRaw]] # comment
        reqFiled filedType [required=true] # commen
'''

class ApiParam(object):
    def __init__(self, lineno, line, name, typ, required, validate, comment):
        self.lineno = lineno
        self.line = line
        self.name = name
        self.typ = typ
        self.required = required
        self.validate = validate
        self.comment = comment
        self.children = []
        self.isNest = False

    def addChildren(self, child):
        assert isinstance(child, ApiParam)
        self.children.append(child)
        return self

    @classmethod
    def fromLine(cls, lineno, line):
        if '#' in line:
            itemsline, comment = line.split('#', 1)
            items = itemsline.split()
        else:
            comment = ''
            items = line.split()
        assert_error(len(items) >= 2, '无效的参数定义', lineno, line)
        name, typ = items[0], items[1]
        assert_error(is_valid_var(name), '无效的参数名定义', lineno, line)
        required = True
        validate = ''
        if len(items) > 2:
            required_validate = ' '.join(items[2:])
            assert_error(required_validate.startswith('false') or required_validate.startswith('true'), '无效的参数定义,校验参数必须以true/false开头', lineno, line)
            required = not required_validate.startswith('false')
            validate = required_validate.lstrip('false:').lstrip('true:')
            if len(validate) > 0:
                assert_error(validate.startswith('"'), '无效的参数定义,校验参数规则必须以"开头', lineno, line)
                validate = validate.lstrip('"')
                vidx = validate.find('"')
                assert_error(vidx != -1, '无效的参数定义,校验参数规则必须以"结尾', lineno, line)
                validate = validate[:vidx]
                # items[2] = required_validate[:vidx+1]

        _isList = False
        # list
        if typ.endswith('[]'):
            _isList = True
            typ = typ[:-2]
        if typ.startswith('[]'):
            _isList = True
            typ = typ[2:]
        isNest = False
        # assert_error(typ in TYPEMAPPING, '无效的参数类型定义', lineno, line)
        if typ in TYPEMAPPING:
            typ = TYPEMAPPING[typ]
            _gotyp = '[]' + typ if _isList else typ
        elif typ.startswith('map[str]'): # map with type struct
            isNest = True
            _gotyp = 'map[string]struct'
        else:
            isNest = True
            _gotyp = '[]struct' if _isList else 'struct'
        p = cls(lineno, line, name, _gotyp, required, validate, comment.strip())
        p.isNest = isNest
        return p
    
    def __repr__(self):
        return '%s %s %s # %s\n' % (self.name, self.typ, self.required, self.comment)

    def code(self):
        validate = ''
        if self.required:
            validate = 'required'
            if len(self.validate) > 0:
                validate += ',%s' % self.validate
        else:
            if len(self.validate) > 0:
                validate = 'omitempty,%s' % self.validate

        tag = '`json:"%s" query:"%s" form:"%s"`' % (self.name, self.name, self.name)
        if len(validate) > 0:
            tag = '`json:"%s" query:"%s" form:"%s" validate:"%s"`' % (self.name, self.name, self.name, validate)

        if self.isNest:
            # Items *[]struct{
            # }
            line = '%s *%s{' % (upperFirst(self.name), self.typ)
            for child in self.children:
                line += child.code()
            line += '} %s // %s\n' % (tag, self.comment)
            return line
        else:
            # * 使用指针类型，由于golang 不支持默认参数，不过可以同nil和非nil 来实现
            line = '%s *%s %s // %s\n' % (upperFirst(self.name), self.typ, tag, self.comment)
            return line

    def schema(self):
        isArray = self.typ.startswith('[]')
        typ = self.typ.lstrip('[]')
        docTyp = OPENAPIDOC_TYPE.get(typ, 'object')
        m = SwaggerSchema(self.name, docTyp, '', self.comment)
        if isArray:
            arr = SwaggerSchema(self.name, docTyp, '', self.comment)

            if len(self.children) > 0:
                for c in self.children:
                    arr.addPropertieItem(c.name, c.schema())
                    if c.required:
                        arr.addRequiredParam(c.name)
            m.setArrayItem(arr)
        else:
            if len(self.children) > 0:
                for c in self.children:
                    m.addPropertieItem(c.name, c.schema())
                    if c.required:
                        m.addRequiredParam(c.name)

        return m


class ApiResp(object):
    def __init__(self, lineno, line, name, typ, required, comment):
        self.lineno = lineno
        self.line = line
        self.name = name
        self.typ = typ
        self.required = required
        self.comment = comment
        self.children = []
        self.isNest = False

    def addChildren(self, child):
        assert isinstance(child, ApiResp)
        self.children.append(child)
        return self

    @classmethod
    def fromLine(cls, lineno, line):
        if '#' in line:
            itemsline, comment = line.split('#', 1)
            items = itemsline.split()
        else:
            comment = ''
            items = line.split()
        assert_error(len(items) >= 2, '无效的参数定义', lineno, line)
        name, typ = items[0], items[1]
        assert_error(is_valid_var(name), '无效的参数名定义', lineno, line)
        required = True
        if len(items) > 2:
            required_validate = ' '.join(items[2:])
            assert_error(required_validate.startswith('false') or required_validate.startswith('true'), '无效的参数定义,校验参数必须以true/false开头', lineno, line)
            required = not required_validate.startswith('false')

        _isList = False
        # list
        if typ.endswith('[]'):
            _isList = True
            typ = typ[:-2]
        if typ.startswith('[]'):
            _isList = True
            typ = typ[2:]
        isNest = False
        # assert_error(typ in TYPEMAPPING, '无效的参数类型定义', lineno, line)
        if typ in TYPEMAPPING:
            typ = TYPEMAPPING[typ]
            _gotyp = '[]' + typ if _isList else typ
        else:
            isNest = True
            _gotyp = '[]struct' if _isList else 'struct'
        p = cls(lineno, line, name, _gotyp, required, comment.strip())
        p.isNest = isNest
        return p
    
    def __repr__(self):
        return '%s %s %s # %s\n' % (self.name, self.typ, self.required, self.comment)

    def code(self):
        tag = ''
        if not self.required:
            tag = '`json:"omitempty,%s"`' % (self.name)
        else:
            tag = '`json:"%s"`' % (self.name)

        if self.isNest:
            line = '%s *%s{' % (upperFirst(self.name), self.typ)
            for child in self.children:
                line += child.code()
            line += '} %s // %s\n' % (tag, self.comment)
            return line
        else:
            # * 使用指针类型，由于golang 不支持默认参数，不过可以同nil和非nil 来实现
            line = '%s *%s %s // %s\n' % (upperFirst(self.name), self.typ, tag, self.comment)
            return line

    def schema(self):
        isArray = self.typ.startswith('[]')
        typ = self.typ.lstrip('[]')
        docTyp = OPENAPIDOC_TYPE.get(typ, 'object')
        m = SwaggerSchema(self.name, docTyp, '', self.comment)
        if isArray:
            arr = SwaggerSchema(self.name, docTyp, '', self.comment)

            if len(self.children) > 0:
                for c in self.children:
                    arr.addPropertieItem(c.name, c.schema())
                    if c.required:
                        arr.addRequiredParam(c.name)
            m.setArrayItem(arr)
        else:
            if len(self.children) > 0:
                for c in self.children:
                    m.addPropertieItem(c.name, c.schema())
                    if c.required:
                        m.addRequiredParam(c.name)
        return m

class ApiDef(object):
    def __init__(self, lineno, line, group, api, method, auth, proto, comment):
        self.lineno = lineno
        self.line = line
        self.group = group
        self.api = api
        self.method = method
        self.auth = auth
        self.proto = proto
        self.comment = comment
        self.apiParams = []
        self.apiResps = []
        self.apiAlias = []

    def __repr__(self):
        txt = '%s %s %s # %s\n' % (self.api, self.method, self.auth, self.comment)
        for param in self.apiParams:
            isinstance(param, ApiParam)
            txt += ' ' * TAB_WIDTH * 2 + str(param)
        return txt
    
    def apiName(self):
        if '.' in self.api:
            _items = self.api.split('.')
            assert_error(len(_items) >= 2, '无效的api名称定义', self.lineno, self.line)
            return ''.join(map(lambda x: upperFirst(x), _items))
        return upperFirst(self.api)

    def handlerName(self):
        n = self.apiName()
        return n + 'Api' # UserCreateApi

    def apiRouterName(self):
        if '.' in self.api:
            _items = self.api.split('.')
            return self.api.replace('.', '/')
        return self.api

    def reqModelName(self):
        return 'Req%s%s' % (upperFirst(self.group), self.apiName())

    def respModelName(self):
        return 'Resp%s%s' % (upperFirst(self.group), self.apiName())

    def apiFullName(self):
        return '%s%s' % (self.group, self.apiName())

    def reqModel(self):
        code = 'type %s struct { // %s\n' % (self.reqModelName(), self.comment)
        for param in self.apiParams:
            isinstance(param, ApiParam)
            code += ' ' * TAB_WIDTH + param.code()
        code += 'Context echo.Context `json:"-"`' # append echo.Context
        code += '}\n\n'
        return code

    def respModel(self):
        code = 'type %s struct { // %s\n' % (self.respModelName(), self.comment)
        for resp in self.apiResps:
            isinstance(resp, ApiResp)
            code += ' ' * TAB_WIDTH + resp.code()
        code += 'Context echo.Context `json:"-"`' # append echo.Context
        code += '}\n\n'
        return code
    
    def apiCode(self):
        if self.proto in ['raw', 'respRaw']:
            handler = '%s(r, nil)' % self.handlerName()
        else:
            handler = '%s(r, resp)' % self.handlerName()

        code = ''

        if self.proto == 'raw':
            code = '''
            func %s(c echo.Context) error {
                r := &%s{}
                r.Context = c
                %s
                return RawOKRequest(c)
            }\n
            ''' % (self.apiName(), self.reqModelName(), handler)
        elif self.proto == 'reqRaw':
            code = '''
            func %s(c echo.Context) error {
                r := &%s{}
                r.Context = c
                resp := make(map[string]interface{})
                %s
                return OKRequestWith(c, resp)
            }\n
            ''' % (self.apiName(), self.reqModelName(), handler)
        elif self.proto == 'respRaw':
            code = '''
            func %s(c echo.Context) error {
                r := &%s{}
                if err := c.Bind(r); err != nil {
                    return err
                }
                if err := c.Validate(r); err != nil {
                    return err
                }
                r.Context = c
                %s
                return RawOKRequest(c)
            }\n
            ''' % (self.apiName(), self.reqModelName(), handler)
        else:
            code = '''
            func %s(c echo.Context) error {
                r := &%s{}
                if err := c.Bind(r); err != nil {
                    return err
                }
                if err := c.Validate(r); err != nil {
                    return err
                }
                r.Context = c
                resp := make(map[string]interface{})
                %s
                return OKRequestWith(c, resp)
            }\n
            ''' % (self.apiName(), self.reqModelName(), handler)

        return code

    def apiCodeV2(self):
        handler = '%s(req, resp)' % self.handlerName()

        code = ''

        if self.proto == 'raw':
            code = '''
            func %s(c echo.Context) error {
                req := &%s{}
                req.Context = c
                resp := &%s{}
                resp.Context = c
                %s
                return RawOKRequest(c)
            }\n
            ''' % (self.apiName(), self.reqModelName(), self.respModelName(), handler)
        elif self.proto == 'reqRaw':
            code = '''
            func %s(c echo.Context) error {
                req := &%s{}
                resp := &%s{}
                resp.Context = c
                %s
                return OKRequestWith(c, resp)
            }\n
            ''' % (self.apiName(), self.reqModelName(), self.respModelName(), handler)
        elif self.proto == 'respRaw':
            code = '''
            func %s(c echo.Context) error {
                req := &%s{}
                if err := c.Bind(req); err != nil {
                    return err
                }
                if err := c.Validate(req); err != nil {
                    return err
                }
                req.Context = c
                resp := &%s{}
                resp.Context = c
                %s
                return RawOKRequest(c)
            }\n
            ''' % (self.apiName(), self.reqModelName(), self.respModelName(), handler)
        else:
            code = '''
            func %s(c echo.Context) error {
                req := &%s{}
                if err := c.Bind(req); err != nil {
                    return err
                }
                if err := c.Validate(req); err != nil {
                    return err
                }
                req.Context = c
                resp := &%s{}
                resp.Context = c
                %s
                return OKRequestWith(c, resp)
            }\n
            ''' % (self.apiName(), self.reqModelName(), self.respModelName(), handler)

        return code

    def apiKitPath(self):
        return 'gopher/app/api/%s/%s_kit.go' % (self.group, self.group)

    def apiRespsV1(self):
        funcKits = GOAPIKITS_CACHE.get(self.apiKitPath(), None)
        if not funcKits:
            return []
        apiFunc = funcKits.funcs.get(self.handlerName(), None)
        if not apiFunc:
            return []
        assert isinstance(apiFunc, GoFunc)
        assert_error(self.handlerName() == apiFunc.name, '未找到函数%s返回值' % self.handlerName(), 0, '')
        return apiFunc.resps

    def apiDoc(self):
        doc = '''
/**
* @api {%s} /%s/%s %s
* @apiGroup %s
* @apiName %s
* @apiDescription %s
''' % (self.method, self.group, self.apiRouterName(), self.apiFullName(), self.group, self.apiFullName(), self.comment)
        if self.auth:
            doc += '* @apiPermission %s\n' % self.auth

        for param in self.apiParams:
            if param.required:
                doc += '* @apiParam {%s} %s %s\n' % (APIDOC_TYPE.get(param.typ, 'Object'), param.name, param.comment)
            else:
                doc += '* @apiParam {%s} [%s] %s\n' % (APIDOC_TYPE.get(param.typ, 'Object'), param.name, param.comment)
        
        funcKits = GOAPIKITS_CACHE.get(self.apiKitPath(), None)
        if funcKits:
            apiFunc = funcKits.funcs.get(self.handlerName(), None)
            if apiFunc:
                assert isinstance(apiFunc, GoFunc)
                assert_error(self.handlerName() == apiFunc.name, '未找到函数%s返回值' % self.handlerName(), 0, '')
                for resp in apiFunc.resps:
                    name, isArray, typ, typObj, comment = resp
                    typ = APIDOC_TYPE.get(typ, 'Object')
                    if isArray:
                        typ = '%s[]' % typ
                    doc += '* @apiSuccess {%s} %s %s\n' % (typ, name, comment + ' ' + typObj)
        doc += '*/\n'
        return doc
    
    @classmethod
    def fromLine(cls, lineno, line, group):
        if '#' in line:
            itemsline, comment = line.split('#', 1)
            items = itemsline.split()
        else:
            comment = ''
            items = line.split()
        assert_error(len(items) in [1, 2, 3, 4], '无效的api定义', lineno, line)
        api, method, auth, proto = items[0], 'GET', None, 'json'
        method = 'GET' if len(items) == 1 or items[1].upper() != 'POST' else 'POST'
        if items[1].upper() == 'ANY':
            method = 'Any'
        if len(items) in [3, 4]:
            auth = items[2]
            if auth == 'none':
                auth = None
            # assert_error(auth in AUTH_TYPE, '只支持%s' % AUTH_TYPE, lineno, line)

        if len(items) == 4:
            proto = items[3]
            assert_error(proto in PROTO_TYPE, '只支持%s' % PROTO_TYPE, lineno, line)

        return cls(lineno, line, group, api, method, auth, proto, comment.strip())

    def addApiParam(self, apiParam):
        assert isinstance(apiParam, ApiParam)
        self.apiParams.append(apiParam)

    def addApiResp(self, apiResp):
        assert isinstance(apiResp, ApiResp)
        self.apiResps.append(apiResp)


class ApiGroup(object):
    def __init__(self, lineno, line, group, comment):
        self.lineno = lineno
        self.line = line
        self.group = group
        self.comment = comment
        self.apis = []

    def __repr__(self):
        txt = '%s // %s\n' % (self.group, self.comment)
        for api in self.apis:
            isinstance(api, ApiDef)
            txt += ' ' * TAB_WIDTH + str(api)
        return txt

    def reqModel(self):
        code = ''
        for api in self.apis:
            isinstance(api, ApiDef)
            code += api.reqModel()
        return code

    def respModel(self):
        code = ''
        for api in self.apis:
            isinstance(api, ApiDef)
            code += api.respModel()
        return code

    def models(self):
        code = ''
        for api in self.apis:
            isinstance(api, ApiDef)
            code += api.reqModel()
            code += api.respModel()
        return code

    def apiCode(self):
        code = '''
        // Code generated by m2c. DO NOT EDIT.
        package %s\n
        import (
            . "%s/ak/helper"
            . "%s/ak/gw/model"
           	. "%s/app/api/%s"
            "github.com/labstack/echo"
        )\n
        ''' % (self.group, MOD, MOD, MOD, self.group)
        for api in self.apis:
            isinstance(api, ApiDef)
            code += api.apiCode()
        return code
    
    def apiCodeV2(self):
        code = '''
        // Code generated by m2c. DO NOT EDIT.
        package %s\n
        import (
            . "%s/ak/helper"
            . "%s/ak/gw/model"
           	. "%s/app/api/%s"
            "github.com/labstack/echo"
        )\n
        ''' % (self.group, MOD, MOD, MOD, self.group)
        for api in self.apis:
            isinstance(api, ApiDef)
            code += api.apiCodeV2()
        return code

    def apiRoute(self):
        code = ''
        for api in self.apis:
            isinstance(api, ApiDef)
            if api.auth:
                code += '%s.%s("/%s", %sApi.%s, authDelegateFunc("%s"))\n' % (self.group, api.method, api.apiRouterName(), self.group, api.apiName(), api.auth)
            else:
                code += '%s.%s("/%s", %sApi.%s)\n' % (self.group, api.method, api.apiRouterName(), self.group, api.apiName())
            
            for apiAlias in api.apiAlias:
                if api.auth:
                    code += 'e.%s("%s", %sApi.%s, authDelegateFunc("%s"))\n' % (api.method, apiAlias, self.group, api.apiName(), api.auth)
                else:
                    code += 'e.%s("%s", %sApi.%s)\n' % (api.method, apiAlias, self.group, api.apiName())
        return code

    def initResps(self):
        if self.apiKitPath() not in GOAPIKITS_CACHE:
            lines = open(self.apiKitPath()).read().splitlines()
            GOAPIKITS_CACHE[self.apiKitPath()] = GoApiKits.fromlines(lines)

    def apiDoc(self):
        if self.apiKitPath() not in GOAPIKITS_CACHE:
            lines = open(self.apiKitPath()).read().splitlines()
            GOAPIKITS_CACHE[self.apiKitPath()] = GoApiKits.fromlines(lines)

        doc = ''
        for api in self.apis:
            doc += api.apiDoc()
        return doc

    def apiKitPath(self):
        return 'gopher/app/api/%s/%s_kit.go' % (self.group, self.group)

    def apiBridgePath(self):
        return 'gopher/ak/gw/bridge/%s/%s.go' % (self.group, self.group)

    @classmethod
    def fromLine(cls, lineno, line):
        if '#' in line:
            itemsline, comment = line.split('#', 1)
            items = itemsline.split()
        else:
            comment = ''
            items = line.split()
        assert_error(len(items) >= 1, '无效的api组定义', lineno, line)
        group = items[0]
        assert_error(is_valid_var(group), '关键字必须是有效字母', lineno, line)
        return cls(lineno, line, group, comment.strip())

    def addApi(self, api):
        assert isinstance(api, ApiDef)
        self.apis.append(api)

class ApiModel(object):
    def __init__(self, root, preifx, version, groups, apiServer='', srvVersion='', title='', description='', authmap={}):
        self.root = root
        self.prefix = preifx
        self.version = version
        self.groups = groups

        self.title = title
        self.description = description
        self.apiServer = apiServer
        self.srvVersion = srvVersion
        self.authmap = authmap

    def __repr__(self):
        txt = 'root %s %s\n' % (self.prefix, self.version)
        for group in self.groups:
            isinstance(group, ApiGroup)
            txt += str(group)
        return txt

    def reqModel(self):
        code = '// Code generated by m2c. DO NOT EDIT.\npackage model\nimport "github.com/labstack/echo"\n'
        for group in self.groups:
            isinstance(group, ApiGroup)
            code += group.reqModel()
            code += group.respModel()
        return code

    def models(self):
        code = '// Code generated by m2c. DO NOT EDIT.\npackage model\nimport "github.com/labstack/echo"\n'
        for group in self.groups:
            isinstance(group, ApiGroup)
            code += group.models()
        return code

    def apiCode(self):
        code = ''
        for group in self.groups:
            isinstance(group, ApiGroup)
            code += group.apiCode()
        return code

    def apiCodeV2(self):
        code = ''
        for group in self.groups:
            isinstance(group, ApiGroup)
            code += group.apiCodeV2()
        return code
    

    def apiRoute(self):
        code = '// Code generated by m2c. DO NOT EDIT.\npackage gw\n'
        code += 'import (\n'
        code += ' ' * TAB_WIDTH + '"github.com/labstack/echo"\n'
        for group in self.groups:
            code += '%sApi "%s/ak/gw/bridge/%s"\n' % (group.group, MOD, group.group)
        code += ')\n'

        code += 'type AuthDelegateFunc func(name string) echo.MiddlewareFunc\n'
        code += 'func LoadServiceApi(e *echo.Echo, authDelegateFunc AuthDelegateFunc) (g *echo.Group) {\n'
        code += 'root := e.Group("/%s/%s")\n\n' % (self.prefix, self.version)
        for group in self.groups:
            code += '%s := root.Group("/%s")\n' % (group.group, group.group)
            code += group.apiRoute()
            code += '\n'
        code += 'return root\n'
        code += '}\n'
        return code

    def initResps(self):
        for group in self.groups:
            group.initResps()

    def apiDoc(self):
        doc = '// Code generated by m2c. DO NOT EDIT.\npackage gw\n'
        doc += '''
 /**
 * @apiDefine userAuth 用户登录
 * 当前接口需要用户登录授权
 */
 
 /**
 * @apiDefine adminAuth 管理员登录
 * 当前接口需要管理员登录授权
 */
 '''
        for group in self.groups:
            doc += group.apiDoc()
            doc += '\n'
        return doc

    def genApiKit(self, group):
        isinstance(group, ApiGroup)
        kitPath = group.apiKitPath()
        code = ''
        if not os.path.exists(kitPath):
            code += 'package %s\n' % group.group
            code += 'import (\n'
            code += '. "%s/ak/gw/model"\n' % MOD
            code += ')\n'
            for api in group.apis:
                code += '// %s\n' % api.comment
                code += 'func %s(req *%s, resp map[string]interface{}) {\n' % (api.handlerName(), api.reqModelName())
                code += 'return\n}\n'
        else:
            # append kit
            oldFuncs = getCodeFuncs(kitPath)
            for api in group.apis:
                if api.handlerName() in oldFuncs:
                    continue

                code += '// %s\n' % api.comment
                code += 'func %s(req *%s, resp map[string]interface{}) {\n' % (api.handlerName(), api.reqModelName())
                code += 'return\n}\n'

        appendCode(kitPath, code)

    def genApiKitV2(self, group):
        isinstance(group, ApiGroup)
        kitPath = group.apiKitPath()
        code = ''
        if not os.path.exists(kitPath):
            code += 'package %s\n' % group.group
            code += 'import (\n'
            code += '. "%s/ak/gw/model"\n' % MOD
            code += ')\n'
            for api in group.apis:
                code += '// %s\n' % api.comment
                code += 'func %s(req *%s, resp *%s) {\n' % (api.handlerName(), api.reqModelName(), api.respModelName())
                code += 'return\n}\n'
        else:
            # append kit
            oldFuncs = getCodeFuncs(kitPath)
            for api in group.apis:
                if api.handlerName() in oldFuncs:
                    continue

                code += '// %s\n' % api.comment
                code += 'func %s(req *%s, resp *%s) {\n' % (api.handlerName(), api.reqModelName(), api.respModelName())
                code += 'return\n}\n'

        appendCode(kitPath, code)

    def genApiCode(self):
        # api model
        modelPath = 'gopher/ak/gw/model/model.go'
        writeCode(modelPath, self.reqModel())

        # api group
        for group in self.groups:
            isinstance(group, ApiGroup)
            code = group.apiCode()
            apiPath = group.apiBridgePath()
            writeCode(apiPath, code)
            self.genApiKit(group)
        
        # route
        routePath = 'gopher/ak/gw/router.go'
        writeCode(routePath, self.apiRoute())

        os.system('gofmt -w gopher/app gopher/ak')

    def genApiCodeV2(self):
        # api model
        modelPath = 'gopher/ak/gw/model/model.go'
        writeCode(modelPath, self.models())

        # api group
        for group in self.groups:
            isinstance(group, ApiGroup)
            code = group.apiCodeV2()
            apiPath = group.apiBridgePath()
            writeCode(apiPath, code)
            self.genApiKitV2(group)
        
        # route
        routePath = 'gopher/ak/gw/router.go'
        writeCode(routePath, self.apiRoute())

        os.system('gofmt -w gopher/app gopher/ak')

    @classmethod
    def fromLines(cls, lines):
        PREFIX = ''
        VERSION = 'v1'
        lineno = 1
        root = lines[0].strip()
        assert_error(is_valid_var(root[0]), '开头必须是有效字母', lineno, root)
        items = root.split(' ')
        assert_error(len(items) >= 2, '无效的apimodel', lineno, root)
        assert_error(items[0] == 'root', '无效的apimodel', lineno, root)
        PREFIX = items[1]
        assert_error(is_valid_var(PREFIX), '关键字必须是有效字母', lineno, root)
        if len(items) == 3:
            VERSION = items[2]
            assert_error(VERSION.startswith('v'), '版本必须以v开头', lineno, root)
            assert_error(VERSION[1:].isdigit(), '版本必须是数字', lineno, root)

        parentDeep = 0
        groups = []
        currentGroup = None
        currentApi = None
        currentNestParam = []
        currentNestResp = []
        inReqScope = True
        inRespScope = False

        title = ''
        description = ''
        apiServer = ''
        srvVersion = ''
        authmap = {}

        for line in lines[1:]:
            lineno += 1

            if not line.strip():
                continue

            # config
            if line.startswith('- title:'):
                title = line.replace('- title:', '').strip()
                continue
            if line.startswith('- description:'):
                description = line.replace('- description:', '').strip()
                continue
            if line.startswith('- apiserver:'):
                apiServer = line.replace('- apiserver:', '').strip()
                continue
            if line.startswith('- version:'):
                srvVersion = line.replace('- version:', '').strip()
                continue
            if line.startswith('- authmap:'):
                line = line.replace('- authmap:', '')
                for item in line.split(','):
                    kv = item.split('=')
                    authmap[kv[0].strip()] = kv[1].strip()
                continue
            
            width = len(line) - len(line.lstrip())
            assert_error(width % TAB_WIDTH == 0, '宽带不是%s的倍数' % TAB_WIDTH, lineno, line)
            deep = int(width / TAB_WIDTH)
            # assert_error(deep <= API_MAX_DEEP_NUM, '深度不能超过%s' % API_MAX_DEEP_NUM, lineno, line)

            line = line.strip()
            # clean nest param
            if deep in [0, 1] and len(currentNestParam) > 0 and currentApi:
                currentNestParam = []
            # clean nest resp
            if deep in [0, 1] and len(currentNestResp) > 0 and currentApi:
                currentNestResp = []

            # group
            if deep == 0:
                currentGroup = ApiGroup.fromLine(lineno, line)
                groups.append(currentGroup)
                inReqScope = False
                inRespScope = False
            # api
            elif deep == 1:
                assert currentGroup
                currentApi = ApiDef.fromLine(lineno, line, currentGroup.group)
                currentGroup.addApi(currentApi)
                inReqScope = False
                inRespScope = False
            # param or resp
            elif deep == 2:
                assert currentApi
                if line.startswith('- resp'):
                    inRespScope = True
                elif line.startswith('- req'):
                    inReqScope = True
                elif line.startswith('- alias:'):
                    apiAlias = line.lstrip('- alias:').strip().split(',')
                    currentApi.apiAlias = apiAlias
                else:
                    if inRespScope:
                        resp = ApiResp.fromLine(lineno, line)
                        if resp.isNest:
                            try:
                                currentNestResp[deep-2] = resp
                            except:
                                currentNestResp.append(resp)
                        currentApi.addApiResp(resp)
                    else:
                        param = ApiParam.fromLine(lineno, line)
                        if param.isNest:
                            try:
                                currentNestParam[deep-2] = param
                            except:
                                currentNestParam.append(param)
                        currentApi.addApiParam(param)
            else:
                if inRespScope:
                    nestResp = currentNestResp[deep-3]
                    assert currentApi
                    resp = ApiResp.fromLine(lineno, line)
                    if resp.isNest:
                        try:
                            currentNestResp[deep-2] = resp
                        except:
                            currentNestResp.append(resp)
                    nestResp.addChildren(resp)
                else:
                    # nest param
                    nestParam = currentNestParam[deep-3]
                    assert currentApi
                    param = ApiParam.fromLine(lineno, line)
                    if param.isNest:
                        try:
                            currentNestParam[deep-2] = param
                        except:
                            currentNestParam.append(param)
                    nestParam.addChildren(param)

        return cls(root, PREFIX, VERSION, groups, apiServer, srvVersion, description, authmap)

class GoFunc(object):
    def __init__(self, name):
        self.name = name
        self.params = []
        self.resps = []

    def __str__(self):
        return '%s - %s - %s' % (self.name, self.params, self.resps)

GOAPIKITS_CACHE = {}
class GoApiKits(object):
    def __init__(self):
        self.funcs = {}
    
    @classmethod
    def fromlines(cls, lines):
        kit = cls()
        current_gofunc = None
        for line in lines:
            # func start
            if line.startswith('func'):
                if 'Api(req' not in line: # not api func
                    continue
                idx = line.find('(')
                funcname = line[:idx].split()[1]

                current_gofunc = GoFunc(funcname)

                left = line[idx+1:]
                idx = left.find(')')
                right = left[:idx]
                for ps in right.split(','):
                    current_gofunc.params.append(set(ps.split()))


            # func end
            if line.startswith('}'):
                if current_gofunc:
                    kit.funcs[current_gofunc.name] = current_gofunc

                current_gofunc = None

            # not in func
            if current_gofunc is None:
                continue

            # func body
            line = line.strip()
            if not line.startswith('resp'):
                continue
            
            comment = ''
            value_type = 'object'
            typObj = ''
            isArray = False

            idx = line.rfind('//')
            if idx != -1:
                # comment for type doc
                comment = line[idx+2:].strip()
                if comment.startswith('@'):
                    typObj = comment.split(' ', 1)[0].strip('@')
                    comment = comment.lstrip('@%s' % typObj).strip()
                elif comment.startswith('object='):
                    typObj = comment.split(' ', 1)[0].lstrip('object=')
                    comment = comment.lstrip('object=%s' % typObj).strip()
                else:
                    vcs = comment.split(' ', 1)
                    if len(vcs) == 2:
                        value_type, comment = vcs
                    elif len(vcs) == 1:
                        if vcs[0] in ['int', 'num', 'number', 'float', 'bool', 'string', 'str']:
                            value_type = vcs[0]
                            comment = ''
                        # value_type = APIDOC_TYPE.get(value_type, 'object')

                if typObj.startswith('[]'):
                    isArray = True
                    typObj = typObj.lstrip('[]')
            
            t = line[5]
            if t == '"':
                # resp["xxx"]
                line = line[6:]
                idx = line.find('"]')
            else:
                # resp[value]
                line = line[5:]
                idx = line.find(']')
            resp = line[:idx]
            current_gofunc.resps.append((resp, isArray, value_type, typObj, comment))
        return kit



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28

import os
import glob
from .apimodel import ApiModel, ApiDef
from .objmodel import ObjModel
from .conf import Color, MOD
from .helper import IS_PY3

# OpenAPI Swargger https://swagger.io/specification/v2

SWAGGER_INDEX = '''
<!-- HTML for static distribution bundle build -->
<!DOCTYPE html>
<html lang="en">
  <head>
    <meta charset="UTF-8">
    <title>Swagger UI</title>
    <link rel="stylesheet" type="text/css" href="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.22.0/swagger-ui.css" >
    <style>
      html
      {
        box-sizing: border-box;
        overflow: -moz-scrollbars-vertical;
        overflow-y: scroll;
      }

      *,
      *:before,
      *:after
      {
        box-sizing: inherit;
      }

      body
      {
        margin:0;
        background: #fafafa;
      }
    </style>
  </head>

  <body>
    <div id="swagger-ui"></div>

    <script src="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.22.0/swagger-ui-bundle.js"> </script>
    <script src="https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.22.0/swagger-ui-standalone-preset.js"> </script>
    <script>
    window.onload = function() {
      // Begin Swagger UI call region
      let hash = window.location.hash
      let url = window.location.href
      url = url.replace(hash, '') + '.yaml'
      const ui = SwaggerUIBundle({
        url: url,
        dom_id: '#swagger-ui',
        deepLinking: true,
        presets: [
          SwaggerUIBundle.presets.apis,
          SwaggerUIStandalonePreset
        ],
        plugins: [
          SwaggerUIBundle.plugins.DownloadUrl
        ],
        layout: "StandaloneLayout"
      })
      // End Swagger UI call region

      window.ui = ui
    }
  </script>
  </body>
</html>
'''

class SwaggerInformation(object):
    def __init__(self, title, description, version):
        self.title = title
        self.description = description
        self.version = version

    def todict(self):
        return {'title': self.title, 'description': self.description, 'version': self.version}

class SwaggerTag(object):
    def __init__(self, name, description):
        self.name = name
        self.description = description

    def todict(self):
        return {'name': self.name, 'description': self.description}

class Swagger(object):
    def __init__(self, host, basePath=''):
        self.swaggerVersion = '2.0'
        self.host = host
        self.basePath = basePath
        self.schemes = ['https']
        self.produces = ['application/json']
        self.tags = [] # groups
        self.paths = {} # apis {"/login", SwaggerItem}
        self.definitions = {} # models, properties_name: properties
        self.info = None

    def goType2SwaggerType(self, typ):
        # array, boolean, integer, null, number, object, string
        if typ.startswith('[]'):
            return 'array'

        if typ.startswith('str'):
            return 'string'

        if typ.startswith('num'):
            return 'number'

        if typ.startswith('int'):
            return 'number'

        if typ.startswith('float'):
            return 'number'

        if typ.startswith('bool'):
            return 'boolean'

        if typ.startswith('map'):
            return 'object'

        return 'object'


    def setInfo(self, title, description, version):
        self.info = SwaggerInformation(title, description, version)

    def addTag(self, name, description):
        self.tags.append(SwaggerTag(name, description))

    def addPath(self, name, sgItem):
        assert isinstance(sgItem, SwaggerItem)
        assert name not in self.paths
        self.paths[name] = sgItem

    def addDefinition(self, name, sgProp):
        assert isinstance(sgProp, SwaggerPropertie)
        assert name not in self.definitions, Color.red('definition "%s" conflict.' % name)
        self.definitions[name] = sgProp

    def todict(self):
        d = {'swagger': self.swaggerVersion, 'host': self.host, 'basePath': self.basePath}
        d['schemes'] = self.schemes
        dps = {}
        for k, v in self.paths.items(): # py2 vs py3 
            dps[k] = v.todict()
        d['paths'] = dps
        dds = {}
        for k, v in self.definitions.items():
            dds[k] = v.todict()
        d['definitions'] = dds
        d['tags'] = list(map(lambda x: x.todict(), self.tags)) # py2 vs py3
        if self.info:
            d['info'] = self.info.todict()
        return d


class SwaggerItem(object):
    def __init(self, postop=None, getop=None):
        self.path = '' # /user/login
        self.postop = postop # post operation
        self.getop = getop # get operation

    def setGetOperation(self, op):
        assert isinstance(op, SwaggerOperation)
        self.getop = op

    def setPostOperation(self, op):
        assert isinstance(op, SwaggerOperation)
        self.postop = op
    
    def todict(self):
        d = {}
        if hasattr(self, 'getop') and self.getop:
            d['get'] = self.getop.todict()
        if hasattr(self, 'postop') and self.postop:
            d['post'] = self.postop.todict()
        return d


class SwaggerOperation(object):
    def __init__(self, summary, description, operationID):
        self.summary = summary # short desc
        self.description = description
        self.operationID = operationID # apiName
        self.tags = [] # apiGroup
        self.consumes = [] # req content-type
        self.produces = [] # resp content-type
        self.parameters = []
        self.responses = []
        # self.security = None

    def addTag(self, tag):
        self.tags.append(tag)

    def addProduce(self, produce):
        self.produces.append(produce)

    def addParam(self, param):
        assert isinstance(param, SwaggerParameter)
        self.parameters.append(param)

    def addResp(self, resp):
        assert isinstance(resp, SwaggerResponse)
        self.responses.append(resp)

    def todict(self):
        d = {'summary': self.summary, 'description': self.description, 'operationId': self.operationID}
        if len(self.tags) > 0:
            d['tags'] = self.tags
        if len(self.consumes) > 0:
            d['consumes'] = self.consumes
        else:
            d['consumes'] = ['application/json']
        if len(self.produces) > 0:
            d['produces'] = self.produces
        else:
            d['produces'] = ['application/json']
        if len(self.parameters) > 0:
            d['parameters'] = list(map(lambda x: x.todict(), self.parameters))
        if len(self.responses) > 0:
            respd = {}
            for resp in self.responses:
                respd[resp.status_code] = resp.todict()
            d['responses'] = respd
        return d


class SwaggerParameter(object):
    def __init__(self, in_, name, description, required=True):
        self.in_ = in_
        self.name = name # param_name
        self.description = description
        self.required = required

        self.schema = None # in body

        self.type = '' # not in body
        self.format = '' # not in body
        self.items = [] # not in body, array item type

        self.default = None

    def setSchema(self, sc):
        assert isinstance(sc, SwaggerSchema)
        self.schema = sc

    def setInQuery(self, type, format):
        self.type = type
        self.format = format

    def addParameterItems(self, item):
        assert isinstance(item, SwaggerParameterItems)
        self.items.append(item)

    def todict(self):
        d = {'in': self.in_, 'name': self.name, 'description': self.description, 'required': self.required}
        if self.in_ == 'body' and self.schema:
            d['schema'] = self.schema.todict()
        else:
            if self.type:
                d['type'] = self.type
            # if self.format:
            #     d['format'] = self.format
            if len(self.items) > 0:
                d['items'] = list(map(lambda x: x.todict(), self.items))
        if self.default:
            d['default'] = self.default
        return d

class SwaggerParameterItems(object):
    def __init__(self, type, format):
        self.type = type
        self.format = format
        self.default = None
        self.items = [] # nest

    def addParameterItems(self, item):
        assert isinstance(item, SwaggerParameterItems)
        self.items.append(item)

    def todict(self):
        d = {}
        if self.type:
            d['type'] = self.type
        # if self.format:
        #     d['format'] = self.format

        if self.default:
            d['default'] = self.default

        if len(self.items) > 0:
            d['items'] = list(map(lambda x: x.todict(), self.items)) # py2 vs py3
        return d


class SwaggerSchema(object):
    def __init__(self, type, ref, title, format, description, ):
        self.type = type # array / object
        self.ref = ref
        self.title = title
        self.format = format
        self.description = description

        self.required = [] # param_name in properties
        self.items = None # type array's item schema
        self.properties = {} # param_name : propertie
        self.enum = [] # options
        self.example = {} # param_name : param_value

    def addRequiredParam(self, name):
        self.required.append(name)

    def addPropItem(self, name, prop):
        assert isinstance(prop, SwaggerPropertie)
        self.properties[name] = prop

    def setArrayItems(self, schemaItem):
        assert isinstance(schemaItem, SwaggerSchema)
        self.type = 'array'
        self.items = schemaItem

    def todict(self):
        d = {}
        if self.ref:
            d['$ref'] = self.ref
        elif self.items:
            d['items'] = self.items.todict()
        else:
            # d.update({'description': self.description})
            d.update({'title': self.title, 'description': self.description})
            if self.type:
                d['type'] = self.type
            # if self.format:
            #     d['format'] = self.format

        if len(self.required) > 0:
            d['required'] = self.required

        dp = {}
        for k, v in self.properties.items():
            dp[k] = v.todict()
        if len(dp) > 0:
            d['properties'] = dp
        if len(self.enum) > 0:
            d['enum'] = self.enum

        if len(self.example) > 0:
            d['example'] = self.example
        return d

class SwaggerPropertie(object):
    def __init__(self, name, type, ref, title, format, description):
        self.name = name
        self.type = type
        self.ref = ref
        self.title = title
        self.format = format
        self.description = description
        self.required = []
        self.default = None
        self.items = None # type == array, set items
        self.additionalProperties = None

        self.properties = {}
    
    def setPropArrayItem(self, item):
        assert isinstance(item, SwaggerPropertie)
        self.items = item

    def addPropertie(self, name, item):
        assert isinstance(item, SwaggerPropertie)
        self.properties[name] = item

    def addRequiredParam(self, name):
        self.required.append(name)

    def todict(self):
        d = {}
        if self.ref:
            d['$ref'] = self.ref
        else:
            d.update({'type': self.type, 'description': self.description})
            # d.update({'type': self.type, 'title': self.title, 'description': self.description})
            # if self.format:
            #     d['format'] = self.format
        if len(self.required) > 0:
            d['required'] = self.required
        if self.default:
            d['default'] = self.default
        if self.items and self.type == 'array':
            d['items'] = self.items.todict()
        if self.additionalProperties:
            d['additionalProperties'] = self.additionalProperties.todict()
        
        dp = {}
        for k, v in self.properties.items():
            dp[k] = v.todict()
        if len(dp) > 0:
            d['properties'] = dp
        return d

class SwaggerResponse(object):
    def __init__(self, description, status_code='200'):
        self.status_code = status_code
        self.description = description
        self.schema = None

    def setSchema(self, sc):
        assert isinstance(sc, SwaggerSchema)
        self.schema = sc

    def todict(self):
        d = {'description': self.description}
        if self.schema:
            d['schema'] = self.schema.todict()
        return d

class GoStruct(object):
    def __init__(self, name, isArray, comment):
        self.name = name
        self.isArray = isArray
        self.comment = comment
        self.fields = []

# apidef -> swagger model def
def _apiParams2SwaggerDef(sg, sgdef, params):
    for pa in params:
        if pa.required:
            sgdef.addRequiredParam(pa.name)

        typ = sg.goType2SwaggerType(pa.typ)
        ref = ''
        # items
        paramProp = SwaggerPropertie(pa.name, typ, ref, pa.name, typ, pa.comment)

        if len(pa.children) > 0:
            childs = SwaggerPropertie('item', 'object', '', '', '', pa.comment)
            # items.item
            childs = _apiParams2SwaggerDef(sg, childs, pa.children)
            paramProp.setPropArrayItem(childs)

        elif typ == 'array':
            itemType = sg.goType2SwaggerType(pa.typ.strip('[]'))
            childs = SwaggerPropertie('item', itemType, '', '', '', pa.comment)
            paramProp.setPropArrayItem(childs)

        # object.field
        sgdef.addPropertie(pa.name, paramProp)
    return sgdef


def _getGoStruct(lines):
    cs = None
    si = {}
    for line in lines:
        # not in type struct
        if not line.strip().startswith('type') and 'struct' not in line and not cs:
            continue
        
        # type struct define
        if line.strip().startswith('type') and 'struct' in line:
            items = line.strip().split(' ')
            assert len(items) >= 4
            cs = items[1]

            si[cs] = []
            continue # next line

        # end struct, do not strip line
        if line.startswith('}'):
            cs = None
            continue # next line

        # struct body
        idx = line.find('//')
        comment = ''
        if idx != -1:
            comment = line[idx+2:].strip()
            line = line[:idx].strip()
        
        typObj = ''
        tag = ''
        idx = line.find('`')
        if idx != -1:
            tag = line[idx+7:line.find('"', idx+7)]
            line = line[:idx].strip()

        # xx string, trick way to split by N space
        items = ' '.join(line.split(' ')).split()

        # case:
        # type Obj struct {
        #     OtherObj
        #     xxx string
        # }
        if len(items) == 0:
            continue

        if len(items) == 1: # object inject
            objname = items[0]
            typ = 'object'
            if objname.startswith('[]'):
                typ = '[]object'

            typObj = objname.lstrip('[]').lstrip('*')
            typObj = typObj.split(".")[-1]
        else:
            typ = items[1]
            if tag == '':
                tag = items[0]

        requried = True
        if 'omitempty' in tag:
            requried = False

        # `name,omitempty xx,xxx`
        idx = tag.find(',')
        if idx != -1:
            tag = tag[:idx]

        isArray = False

        if typ.startswith('[]'):
            isArray = True
            typ = typ.lstrip('[]')

        if typ.startswith('*'):
            typ = typ.lstrip('*')

        if len(items) > 1 and typ not in ['string', 'int64', 'bool']: # not in base type
            typObj = typ
            typ = 'Object'
            # comment for type doc
            # comment = line[idx+2:].strip()
            # if comment.startswith('[]'):
            #     isArray = True
            #     comment = comment.lstrip('[]')

            if comment.startswith('@'):
                typObj = comment.split(' ')[0].lstrip('@')
                comment = comment.lstrip('@%s' % typObj).strip()
            elif comment.startswith('object='):
                typObj = comment.split(' ')[0].lstrip('object=')
                comment = comment.lstrip('object=%s' % typObj).strip()

            if typObj.startswith('[]'):
                isArray = True
                typObj = typObj.lstrip('[]')

        i = (requried, tag, isArray, typ, typObj, comment)

        si[cs].append(i)
    return si
        


def genSwaggerContent(url, title, description, version):
    # sg
    sg = Swagger(url)
    sg.setInfo(title, description, version)

    # user defines
    infoPath = 'gopher/app/info/info.go'
    lines = open(infoPath).read().splitlines()
    ds = _getGoStruct(lines)
    for item in ds.items(): # py2 vs py3
        name, props = item
        propItem = SwaggerPropertie(name, 'object', '', name, 'object', name)

        for prop in props:
            requried, param, isArray, typ, refObj, comment = prop
            if requried:
                propItem.addRequiredParam(param)

            typ = sg.goType2SwaggerType(typ)

            ref = ''
            if refObj:
                typ = 'object'
                ref = '#/definitions/%s' % refObj

            paramProp = SwaggerPropertie(param, typ, ref, name, typ, comment)
            if isArray:
                swprArray = SwaggerPropertie('', 'array', '', '', 'array', comment) # here ref need stuff
                swprArray.setPropArrayItem(paramProp)
                propItem.addPropertie(param, swprArray)
            else:
                propItem.addPropertie(param, paramProp)

        sg.addDefinition(name, propItem)
    
    # model defines
    lines = open('objmodel.txt', 'rb').readlines()
    if IS_PY3(): # py2 vs py3
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    objModel = ObjModel.fromLines(lines)
    objModel.checkObjCode() 
    # obj
    for topobj in objModel.root.children:
        objs = topobj.getObjs()
        for obj in objs:
            objname, comment, objfields = obj
            propItem = SwaggerPropertie(objname, 'object', '', objname, 'object', comment)

            for prop in objfields:
                param, isArray, typ, comment = prop
                propItem.addRequiredParam(param)

                ref = ''
                typ = sg.goType2SwaggerType(typ)

                paramProp = SwaggerPropertie(param, typ, ref, param, typ, comment)
                propItem.addPropertie(param, paramProp)

            sg.addDefinition(objname, propItem)


    # api
    lines = open('apimodel.txt', 'rb').readlines()
    if IS_PY3(): # py2 vs py3
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    apiModel = ApiModel.fromLines(lines)
    apiModel.initResps()
    for group in apiModel.groups:
        sg.addTag(group.group, group.comment)

        for api in group.apis:
            path = SwaggerItem()
            op = SwaggerOperation(api.apiName(), api.comment, api.apiFullName())
            if api.proto == 'raw':
                op.addProduce('application/octet-stream')
            else:
                op.addProduce('application/json')

            op.addTag(api.group)

            # api req model
            propItem = SwaggerPropertie(api.reqModelName(), 'object', '', api.reqModelName(), 'object', api.comment)
            propItem = _apiParams2SwaggerDef(sg, propItem, api.apiParams)
            sg.addDefinition(api.reqModelName(), propItem)

            # request param
            paramIn = 'query' if api.method.lower() == 'get' else 'body'
            if paramIn == 'body':
                swpa = SwaggerParameter(paramIn, 'body', api.comment, True) 
                swpa.schema = SwaggerSchema('object', '#/definitions/%s' % api.reqModelName(), api.apiFullName(), '', api.comment)
                op.addParam(swpa)
            else:
                for pa in api.apiParams:
                    swpa = SwaggerParameter(paramIn, pa.name, pa.comment, pa.required)
                    swpa.setInQuery(sg.goType2SwaggerType(pa.typ), pa.typ)
                    op.addParam(swpa)

            # response
            scrp = SwaggerSchema('object', '', api.apiFullName(), '', api.comment)
            for rp in api.apiRespsV1():
                name, isArray, typ, typObj, comment = rp 
                scrp.addRequiredParam(name)

                typ = sg.goType2SwaggerType(typ)
                ref = ''
                if typObj:
                    ref = '#/definitions/%s' % typObj
                    typ = 'object'

                # single object or array object
                swpr = SwaggerPropertie(name, typ, ref, name, typ, comment) # here ref need stuff
                if not isArray:
                    scrp.addPropItem(name, swpr)
                else:
                    swprArray = SwaggerPropertie('', 'array', '', '', 'array', comment) # here ref need stuff
                    swprArray.setPropArrayItem(swpr)
                    scrp.addPropItem(name, swprArray)

            swrp = SwaggerResponse(api.comment)
            swrp.setSchema(scrp)
            op.addResp(swrp)

            if api.method.lower() == 'get':
                path.setGetOperation(op)
            else:
                path.setPostOperation(op)

            apiName = '/api/%s/%s/%s' % (apiModel.version, api.group, api.apiRouterName())
            sg.addPath(apiName, path)
    

    d = sg.todict()
    from yaml import safe_dump
    return safe_dump(d)


def genSwaggerYaml(savePath, url, title, description, version):
    content = genSwaggerContent(url, title, description, version)
    fp = open(savePath, 'wb')
    header = '# Code generated by m2c. DO NOT EDIT.\n'
    if IS_PY3():
        fp.write(header.encode('utf-8'))
        fp.write(content.encode('utf-8'))
    else:
        fp.write(header)
        fp.write(content)
    fp.close()


def genSwaggerGwGo(savePath, url, title, description, version):
    save_path = 'gopher/ak/gw/swagger.go'
    content = genSwaggerContent(url, title, description, version)
    fp = open(savePath, 'wb')
    code = '''
// Code generated by m2c. DO NOT EDIT.
package gw

const SWAGGER_INDEX = `
%s
`

const SWAGGER_CONTENT = `
%s
`
    ''' % (SWAGGER_INDEX, content)
    if IS_PY3():
        fp.write(code.encode('utf-8'))
    else:
        fp.write(code)
    fp.close()
    os.system('gofmt -w gopher/ak/gw')

def genSwaggerGwGoV4(savePath, url, title, description, version):
    save_path = 'gopher/ak/gw/swagger.go'
    content = genSwaggerContentV4(url, title, description, version)
    fp = open(savePath, 'wb')
    code = '''
// Code generated by m2c. DO NOT EDIT.
package gw

const SWAGGER_INDEX = `
%s
`

const SWAGGER_CONTENT = `
%s
`
    ''' % (SWAGGER_INDEX, content)
    if IS_PY3():
        fp.write(code.encode('utf-8'))
    else:
        fp.write(code)
    fp.close()
    os.system('gofmt -w gopher/ak/gw')


def genSwaggerContentV4(url, title, description, version):
    # sg
    sg = Swagger(url)
    sg.setInfo(title, description, version)

    # user defines
    ds = {}
    for ip in glob.glob('%s/model/*.go' % MOD):
        lines = open(ip).read().splitlines()
        ds.update(_getGoStruct(lines))

    for item in ds.items(): # py2 vs py3
        name, props = item
        propItem = SwaggerPropertie(name, 'object', '', name, 'object', name)

        for prop in props:
            requried, param, isArray, typ, refObj, comment = prop
            if requried:
                propItem.addRequiredParam(param)

            typ = sg.goType2SwaggerType(typ)

            ref = ''
            if refObj:
                typ = 'object'
                ref = '#/definitions/%s' % refObj

            paramProp = SwaggerPropertie(param, typ, ref, name, typ, comment)
            if isArray:
                swprArray = SwaggerPropertie('', 'array', '', '', 'array', comment) # here ref need stuff
                swprArray.setPropArrayItem(paramProp)
                propItem.addPropertie(param, swprArray)
            else:
                propItem.addPropertie(param, paramProp)

        sg.addDefinition(name, propItem)
    
    # model defines
    lines = open('objmodel.txt', 'rb').readlines()
    if IS_PY3(): # py2 vs py3
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    objModel = ObjModel.fromLines(lines)
    objModel.checkObjCode() 
    # obj
    for topobj in objModel.root.children:
        objs = topobj.getObjs()
        for obj in objs:
            objname, comment, objfields = obj
            propItem = SwaggerPropertie(objname, 'object', '', objname, 'object', comment)

            for prop in objfields:
                param, isArray, typ, comment = prop
                propItem.addRequiredParam(param)

                ref = ''
                typ = sg.goType2SwaggerType(typ)

                paramProp = SwaggerPropertie(param, typ, ref, param, typ, comment)
                propItem.addPropertie(param, paramProp)

            sg.addDefinition(objname, propItem)


    # api
    lines = open('apimodel.txt', 'rb').readlines()
    if IS_PY3(): # py2 vs py3
        lines = list(map(lambda x: x.decode('utf-8'), lines))
    apiModel = ApiModel.fromLines(lines)
    apiModel.initResps()
    for group in apiModel.groups:
        sg.addTag(group.group, group.comment)

        for api in group.apis:
            path = SwaggerItem()
            op = SwaggerOperation(api.apiName(), api.comment, api.apiFullName())
            if api.proto == 'raw':
                op.addProduce('application/octet-stream')
            else:
                op.addProduce('application/json')

            op.addTag(api.group)

            # api req model
            propItem = SwaggerPropertie(api.reqModelName(), 'object', '', api.reqModelName(), 'object', api.comment)
            propItem = _apiParams2SwaggerDef(sg, propItem, api.apiParams)
            sg.addDefinition(api.reqModelName(), propItem)

            # request param
            paramIn = 'query' if api.method.lower() == 'get' else 'body'
            if paramIn == 'body':
                swpa = SwaggerParameter(paramIn, 'body', api.comment, True) 
                swpa.schema = SwaggerSchema('object', '#/definitions/%s' % api.reqModelName(), api.apiFullName(), '', api.comment)
                op.addParam(swpa)
            else:
                for pa in api.apiParams:
                    swpa = SwaggerParameter(paramIn, pa.name, pa.comment, pa.required)
                    swpa.setInQuery(sg.goType2SwaggerType(pa.typ), pa.typ)
                    op.addParam(swpa)

            # response
            scrp = SwaggerSchema('object', '', api.apiFullName(), '', api.comment)
            for rp in api.apiRespsV1():
                name, isArray, typ, typObj, comment = rp 
                scrp.addRequiredParam(name)

                typ = sg.goType2SwaggerType(typ)
                ref = ''
                if typObj:
                    ref = '#/definitions/%s' % typObj
                    typ = 'object'

                # single object or array object
                swpr = SwaggerPropertie(name, typ, ref, name, typ, comment) # here ref need stuff
                if not isArray:
                    scrp.addPropItem(name, swpr)
                else:
                    swprArray = SwaggerPropertie('', 'array', '', '', 'array', comment) # here ref need stuff
                    swprArray.setPropArrayItem(swpr)
                    scrp.addPropItem(name, swprArray)

            swrp = SwaggerResponse(api.comment)
            swrp.setSchema(scrp)
            op.addResp(swrp)

            if api.method.lower() == 'get':
                path.setGetOperation(op)
            else:
                path.setPostOperation(op)

            apiName = '/api/%s/%s/%s' % (apiModel.version, api.group, api.apiRouterName())
            sg.addPath(apiName, path)
    

    d = sg.todict()
    from yaml import safe_dump
    return safe_dump(d)



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28
from __future__ import print_function
import os
import sys
import json
import platform

def IS_PY3():
    return sys.version_info >= (3, 0)

def assert_error(exp, msg, lineno, line):
    if exp:
        return
    err = 'line(%s): %s, error: %s' % (lineno, line, msg)
    raise Exception(err)

def writeCode(path, code):
    if os.path.dirname(path) != '':
        os.system('mkdir -p %s' % os.path.dirname(path))
    if IS_PY3():
        code = code.encode('utf-8')
    with open(path, 'wb') as f:
        f.write(code)

def appendCode(path, code):
    os.system('mkdir -p %s' % os.path.dirname(path))
    if IS_PY3():
        code = code.encode('utf-8')
    with open(path, 'ab') as f:
        f.write(code)

def getCodeFuncs(path):
    funcs = []
    for line in open(path, 'rb'):
        if IS_PY3():
            line = line.decode('utf-8')
        line = line.strip()
        if not line.startswith('func '):
            continue
        
        item, _ = line.split('(', 1)
        items = item.split()
        if len(items) != 2:
            print('无效的函数定义:%s' % line)
            continue
        funcs.append(items[1])
    return funcs

def upperFirst(s):
    if len(s) in (0, 1):
        return s.upper()
    if '_' in s:
        items = s.split('_')
        name = ''
        for item in items:
            name += upperFirst(item)
        return name
    return s[0].upper() + s[1:]

def is_valid_var(s):
    if len(s) == 0:
        return False
    if s[0].isdigit():
        return False
    s = s.replace('_', '')
    return s.isalnum()

def is_m2c():
    return os.path.exists('.m2c')

def m2c_version():
    if not is_m2c():
        return None
    
    d = json.loads(open('.m2c', 'rb').read())
    return d['__version__']

def m2c_cliversion():
    if not is_m2c():
        return None
    
    d = json.loads(open('.m2c', 'rb').read())
    return d.get('__cliversion__', 'v2')

def update_m2c(new_version, cli_version='v2'):
    d = json.loads(open('.m2c', 'rb').read())
    d['__version__'] = new_version
    d['__cliversion__'] = cli_version
    if IS_PY3():
        open('.m2c', 'wb').write(json.dumps(d, indent=2).encode('utf-8'))
    else:
        open('.m2c', 'wb').write(json.dumps(d, indent=2))

def is_macos():
    return platform.system() == 'Darwin'

def sed_file(path, src_line, dst_line):
    txt = open(path).read()
    txt_dst = ''
    for line in open(path):
        if line == src_line:
            line = dst_line
        txt_dst += line
    writeCode(path, txt_dst)



import setuptools
from m2c.version import version

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="m2c",
    version=version,
    author="yangbinnnn",
    author_email="yangbinnnn@gmail.com",
    description="m2c",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yangbinnnn/m2c",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "pyyaml",
        "psutil",
        "watchdog"
    ],
    scripts=['bin/m2c'],
)



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28
from __future__ import print_function
import os

from .conf import *
from .helper import *
from .objtpl import *

# Object Model to Code
'''
root object # 开始标记
topObjName TopObjName[] # comment, Top Object
    filedObjName [Prefix]filedObjName[Suffix] [*] # comment Filed Object
'''

class ObjNode(object):
    def __init__(self, lineno, line, fieldName, fieldType, containerType, refType, needDump, comment, parent):
        self.lineno = lineno
        self.line = line
        self.fieldName = fieldName
        self.fieldType = fieldType
        self.containerType = containerType
        self.refType = refType
        self.needDump = needDump
        self.comment = comment
        self.parent = parent
        self.children = []

    def addChildren(self, child):
        assert isinstance(child, ObjNode)
        self.children.append(child)
        return self

    def isRootObj(self):
        return self.lineno == 1

    def isBasicType(self):
        return self.containerType == OBJ_CONTAINER_TYPE_BASIC

    def isMgrType(self):
        return self.refType == OBJ_REF_TYPE_MANAGE
    
    def isRefType(self):
        return self.refType == OBJ_REF_TYPE_REF

    def isSingleType(self):
        return self.containerType == OBJ_CONTAINER_TYPE_SINGLE

    def isListType(self):
        return self.containerType == OBJ_CONTAINER_TYPE_LIST

    def longNamespace(self):
        if self.parent is None:
            return None
        if self.containerType == OBJ_CONTAINER_TYPE_BASIC:
            return None
        if self.refType == OBJ_REF_TYPE_REF:
            return None
        pn = self.parent.longNamespace()
        if pn is None:
            return self.fieldType[0].lower() + self.fieldType[1:]
        return pn[0].lower() + pn[1:] + self.fieldType

    def bigLongNamespace(self):
        return upperFirst(self.longNamespace())

    def shortNamespace(self):
        return self.fieldName

    def bigShortNamespace(self):
        return upperFirst(self.shortNamespace())

    def genObjFuncParam(self):
        pm = ''
        for child in self.children:
            if not child.isBasicType():
                continue
            pm += '%s %s,' % (child.fieldName, child.fieldType)
        return pm.strip(',')

    def show(self, ident):
        s = 4 * ident * ' '
        if self.lineno == 0:
            s = self.line
        fieldType = self.fieldType
        if self.containerType == OBJ_CONTAINER_TYPE_LIST:
            fieldType += '[]'
        if self.refType == OBJ_REF_TYPE_REF:
            fieldType = '@' + fieldType
        s += '%s %s %s # %s %s\n' % (self.fieldName, fieldType, self.needDump, self.longNamespace(), self.comment)
        for child in self.children:
            s += child.show(ident+1)
        return s
    # import and define
    def objCode0(self):
        code = '// Code generated by m2c. DO NOT EDIT.\npackage %s\n' % self.fieldName
        code += 'import (\n'
        code += '"time"\n'
        code += '"github.com/globalsign/mgo"\n'
        code += '"github.com/globalsign/mgo/bson"\n'
        code += 'id "%s/ak/orm/id"\n' % MOD
        for child in self.children:
            assert isinstance(child, ObjNode)
            if child.isBasicType(): # skip basic type
                continue
            if child.isRefType():
                code += '%sObject "%s"\n' % (child.shortNamespace(), child.objImportPath())
            else:
                code += '%sObject "%s"\n' % (child.longNamespace(), child.objImportPath())
        code += ')\n'

        code += 'var (\n'
        code += '%sTable *mgo.Collection\n' % self.longNamespace()
        code += 'db      *mgo.Database\n'
        code += ')\n'

        code += 'func Init(mgo *mgo.Database) {\n'
        code += 'db = mgo\n'
        code += '%sTable = mgo.C("%s")\n' % (self.longNamespace(), self.longNamespace())
        for child in self.children:
            assert isinstance(child, ObjNode)
            if child.isBasicType() or child.isRefType():
                continue
            code += '%sObject.Init(db)\n' % child.longNamespace()
        code += '}\n'

        code += 'type %s struct {\n' % self.bigLongNamespace()
        code += '// auto bind\n'
        code += 'Id    int64 `json:"id" bson:"_id"`\n'
        code += 'CTime int64 `json:"ctime" bson:"ctime"`\n'
        code += 'MTime int64 `json:"mtime" bson:"mtime"`\n'

        code += '// basic type\n'
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isBasicType(): # basic type
                continue
            jtag = 'json:"%s"' % ('-' if not child.needDump else child.fieldName)
            code += '%s %s `%s bson:"%s"` // %s\n' % (upperFirst(child.fieldName), child.fieldType, jtag, child.fieldName, child.comment)

        code += '// relation\n'
        if self.isMgrType() and not self.parent.isRootObj(): # 如果当前为manager字段，需要引用parent的ID, 且非数组类型
            # print self.parent.lineno, self.parent.isBasicType(), self.fieldName, self.fieldType
            jtag = 'json:"%s"' % ('-' if not self.parent.needDump else '%sId' % self.parent.longNamespace())
            code += 'Ref%sId int64 `%s bson:"%sId"` // %s\n'  % (self.parent.bigLongNamespace(), jtag, self.parent.longNamespace(), self.comment)

        for child in self.children:
            assert isinstance(child, ObjNode)
            if child.isBasicType() or child.isListType(): # skip basic type and list type
                continue
            name = '%s%sId' % ('Mgr' if child.isMgrType() else 'Ref', upperFirst(child.fieldName))
            jtag = 'json:"%s"' % ('-' if not child.needDump else '%sId' % child.fieldName)
            code += '%s int64 `%s bson:"%sId"` // %s\n' % (name, jtag, child.fieldName, child.comment)
        code += '}\n'

        code += 'func New%s(' % self.bigLongNamespace()
        code += self.genObjFuncParam()
        code += ')*%s {\n' % self.bigLongNamespace()
        code += 'return &%s{\n' % self.bigLongNamespace()
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isBasicType(): # basic type
                continue
            code += '%s: %s,\n' % (child.bigShortNamespace(), child.fieldName)
        code += '}\n}\n'
        return code
        
    # import and define
    def objmodel(self):
        code = '// Code generated by m2c. DO NOT EDIT.\npackage db\n'

        code += 'type %sModel struct {\n' % self.bigLongNamespace()
        code += '// auto bind\n'
        code += 'Id    int64 `json:"id" bson:"_id"`\n'
        code += 'CTime int64 `json:"ctime" bson:"ctime"`\n'
        code += 'MTime int64 `json:"mtime" bson:"mtime"`\n'

        code += '// basic type\n'
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isBasicType(): # basic type
                continue
            jtag = 'json:"%s"' % ('-' if not child.needDump else child.fieldName)
            code += '%s %s `%s bson:"%s"` // %s\n' % (upperFirst(child.fieldName), child.fieldType, jtag, child.fieldName, child.comment)

        code += '// relation\n'
        if self.isMgrType() and not self.parent.isRootObj(): # 如果当前为manager字段，需要引用parent的ID, 且非数组类型
            # print self.parent.lineno, self.parent.isBasicType(), self.fieldName, self.fieldType
            jtag = 'json:"%s"' % ('-' if not self.parent.needDump else '%sId' % self.parent.longNamespace())
            code += 'Ref%sId int64 `%s bson:"%sId"` // %s\n'  % (self.parent.bigLongNamespace(), jtag, self.parent.longNamespace(), self.comment)

        for child in self.children:
            assert isinstance(child, ObjNode)
            if child.isBasicType() or child.isListType(): # skip basic type and list type
                continue
            name = '%s%sId' % ('Mgr' if child.isMgrType() else 'Ref', upperFirst(child.fieldName))
            jtag = 'json:"%s"' % ('-' if not child.needDump else '%sId' % child.fieldName)
            code += '%s int64 `%s bson:"%sId"` // %s\n' % (name, jtag, child.fieldName, child.comment)
        code += '}\n'

        code += 'func New%sModel (' % self.bigLongNamespace()
        code += self.genObjFuncParam()
        code += ')*%sModel {\n' % self.bigLongNamespace()
        code += 'return &%sModel{\n' % self.bigLongNamespace()
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isBasicType(): # basic type
                continue
            code += '%s: %s,\n' % (child.bigShortNamespace(), child.fieldName)
        code += '}\n}\n'
        return code

    def objmodelV4(self):
        code = '// Code generated by m2c. DO NOT EDIT.\npackage db\n'

        code += 'type %sModel struct {\n' % self.bigLongNamespace()
        code += '// auto bind\n'
        code += 'Id    interface{} `json:"id" bson:"_id,omitempty"` // int or string \n'
        code += 'CTime int64 `json:"ctime" bson:"ctime"`\n'
        code += 'MTime int64 `json:"mtime" bson:"mtime"`\n'

        code += '// basic type\n'
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isBasicType(): # basic type
                continue
            jtag = 'json:"%s"' % ('-' if not child.needDump else child.fieldName)
            code += '%s %s `%s bson:"%s"` // %s\n' % (upperFirst(child.fieldName), child.fieldType, jtag, child.fieldName, child.comment)

        for child in self.children:
            assert isinstance(child, ObjNode)
            if child.isBasicType() or child.isListType(): # skip basic type and list type
                continue
            if child.isRefType():
                name = '%s%sId' % ('Ref', upperFirst(child.fieldName))
                jtag = 'json:"%s"' % ('-' if not child.needDump else '%sId' % child.fieldName)
                code += '%s int64 `%s bson:"%sId"` // %s\n' % (name, jtag, child.fieldName, child.comment)
            else: # mangerType
                jtag = 'json:"%s"' % ('-' if not child.needDump else '%s' % child.fieldName)
                code += '%s *%sModel `%s bson:"%s"` // %s\n' % (upperFirst(child.fieldName), upperFirst(child.bigLongNamespace()), jtag, child.fieldName, child.comment)
        code += '}\n'

        code += 'func New%sModel (' % self.bigLongNamespace()
        code += self.genObjFuncParam()
        code += ')*%sModel {\n' % self.bigLongNamespace()
        code += 'return &%sModel{\n' % self.bigLongNamespace()
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isBasicType(): # basic type
                continue
            code += '%s: %s,\n' % (child.bigShortNamespace(), child.fieldName)
        code += '}\n}\n'

        code += self.objmodelV4Mgr()
        return code

    def objmodelV4Mgr(self):
        code = ''
        for child in self.children:
            if not child.isMgrType():
                continue

            code += 'type %sModel struct {\n' % child.bigLongNamespace()
            code += '// basic type\n'
            for subchild in child.children:
                assert isinstance(subchild, ObjNode)
                if not subchild.isBasicType(): # basic type
                    continue
                jtag = 'json:"%s"' % ('-' if not subchild.needDump else subchild.fieldName)
                code += '%s %s `%s bson:"%s"` // %s\n' % (upperFirst(subchild.fieldName), subchild.fieldType, jtag, subchild.fieldName, subchild.comment)

            for subchild in child.children:
                assert isinstance(subchild, ObjNode)
                if subchild.isBasicType() or subchild.isListType(): # skip basic type and list type
                    continue

                jtag = 'json:"%s"' % ('-' if not subchild.needDump else subchild.fieldName)
                code += '%s *%sModel `%s bson:"%s"` // %s\n' % (upperFirst(subchild.fieldName), upperFirst(subchild.bigLongNamespace()), jtag, subchild.fieldName, subchild.comment)
            code += '}\n'

            code += 'func New%sModel (' % child.bigLongNamespace()
            code += child.genObjFuncParam()
            code += ')*%sModel {\n' % child.bigLongNamespace()
            code += 'return &%sModel{\n' % child.bigLongNamespace()
            for subchild in child.children:
                assert isinstance(subchild, ObjNode)
                if not subchild.isBasicType(): # basic type
                    continue
                code += '%s: %s,\n' % (subchild.bigShortNamespace(), subchild.fieldName)
            code += '}\n}\n'

            code += child.objmodelV4Mgr()
        return code

    # fixed code by tpl
    def objCode1(self):
        name = self.longNamespace()
        bigname = name[0].upper() + name[1:]
        code = OBJ_FIXED_CODE_TPL.replace('$BIGOBJNAME$', bigname)
        code = code.replace('$OBJNAME$', name)
        return code

    # helper func for relation
    def objCode2(self):
        code = ''
        name = self.longNamespace()
        bigname = upperFirst(name)
        for child in self.children:
            assert isinstance(child, ObjNode)
            code + '\n'
            fieldName = child.fieldName if child.isBasicType() else '%sId' % child.fieldName
            fieldType = child.fieldType if child.isBasicType() else 'int64'
            values = {'objName': name, 'ObjName': bigname, 'fieldName': fieldName, 'FieldName': upperFirst(fieldName), 'filedType': fieldType}
            code += OBJ_FILTER_HELPER_CODE_TPL.format(**values)
            code + '\n'

            # string type add like
            if fieldType == 'string':
                code += OBJ_FILTER_STRING_LIKE_CODE_TPL.format(**values)
                code += '\n'

        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isMgrType(): 
                continue
            code += '\n'
            # print child.fieldName, child.fieldType, child.lineno
            values = {'objName': name, 'ObjName': bigname, 'mgrShortName': child.shortNamespace(), 'MgrShortName': child.bigShortNamespace(), 'mgrLongName': child.longNamespace(), 'MgrLongName': child.bigLongNamespace()}
            # print child.fieldName, child.fieldType
            if child.isListType():
                code += OBJ_MGR_ARRAY_HELPER_CODE_TPL.format(**values)
            else:
                code += OBJ_MGR_HELPER_CODE_TPL.format(**values)
            code += '\n'

        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isRefType(): 
                continue
            code += '\n'
            values = {'objName': name, 'ObjName': bigname, 'refName': child.shortNamespace(), 'RefName': child.bigShortNamespace()}
            code += OBJ_REF_HELPER_CODE_TPL.format(**values)
            code += '\n'

        if self.isMgrType() and not self.parent.isRootObj():
            code += '\n'
            values = {'objName': name, 'ObjName': bigname, 'parentName': self.parent.longNamespace(), 'ParentName': self.parent.bigLongNamespace()}
            code += OBJ_MGR_REF_HELPER_CODE_TPL.format(**values)
            code += '\n'
        return code

    def objCodeFull(self):
        if self.isBasicType() or self.isRefType():
            return None
        return self.objCode0() + self.objCode1() + self.objCode2()

    def objCodeFullV4(self):
        return self.objmodelV4()

    def objCodePath(self):
        if self.isBasicType() or self.isRefType():
            return None
        p = '%s/%s.go' % (self.fieldName, self.fieldName)
        parent = self.parent
        while parent:
            if parent.lineno == 1:
                break
            assert isinstance(parent, ObjNode)
            p = os.path.join(parent.fieldName, p)
            parent = parent.parent
        return 'gopher/ak/orm/%s' % p

    def objCodePathV4(self):
        if self.isBasicType() or self.isRefType():
            return None
        p = '%s.go' % (self.fieldName)
        parent = self.parent
        while parent:
            if parent.lineno == 1:
                break
            assert isinstance(parent, ObjNode)
            p = parent.fieldName + '_' + p
            parent = parent.parent
        return '%s/model/db/%s' % (MOD, p)
    
    def objImportPath(self):
        p = self.fieldName
        if not self.isRefType():
            parent = self.parent
            while parent:
                if parent.lineno == 1:
                    break
                assert isinstance(parent, ObjNode)
                p = os.path.join(parent.fieldName, p)
                parent = parent.parent
        return 'gopher/ak/orm/%s' % p

    def getObjs(self):
        if self.isBasicType() or self.isRefType():
            return []

        # real object
        objs = []
        objFields = [('id', False, 'int64', '对象ID'), ('ctime', False, 'int64', '对象创建时间'), ('mtime', False, 'int64', '对象修改时间')]

        # basic type
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isBasicType(): # basic type
                continue
            
            if not child.needDump:
                continue
            objFields.append((child.fieldName, child.isListType(), child.fieldType, child.comment))

        # relation fields
        if self.isMgrType() and not self.parent.isRootObj(): # 如果当前为manager字段，需要引用parent的ID
            if self.parent.needDump:
                objFields.append(('%sId' % self.parent.longNamespace(), False, 'int64', self.parent.comment))

        for child in self.children:
            assert isinstance(child, ObjNode)
            if child.isBasicType(): # skip basic type
                continue
            if not child.needDump:
                continue
            name = '%s%sId' % ('Mgr' if child.isMgrType() else 'Ref', upperFirst(child.fieldName))
            objFields.append(('%sId' % child.fieldName, False, 'int64', child.comment))

        obj = (self.bigLongNamespace(), self.comment, objFields)
        objs.append(obj)

        # nest parser child mgr type
        for child in self.children:
            assert isinstance(child, ObjNode)
            if not child.isMgrType():
                continue
            objs.extend(child.getObjs())

        return objs
    
    def genObjCode(self):
        codePath = self.objCodePath()
        if codePath:
            print(codePath)
            code = self.objCodeFull()
            writeCode(codePath, code)

        for child in self.children:
            child.genObjCode()

    def genObjCodeV4(self):
        codePath = self.objCodePathV4()
        if codePath:
            print(codePath)
            code = self.objCodeFullV4()
            writeCode(codePath, code)

        # for child in self.children:
        #     child.genObjCodeV4()

    def checkObjCode(self, topObjects):
        for child in self.children:
            if child.isRefType():
                assert_error(child.fieldType in topObjects, '引用对象%s未定义' % child.fieldType, child.lineno, child.line)
            child.checkObjCode(topObjects)

    @classmethod
    def fromLine(cls, lineno, line, parent):
        assert isinstance(parent, ObjNode)
        if '#' in line:
            itemsline, comment = line.split('#', 1)
            items = itemsline.split()
        else:
            comment = ''
            items = line.split()
        assert_error(len(items) in (2, 3), '无效的object定义', lineno, line)
        needDump = True
        if len(items) == 3:
            needDump = False
        fieldName = items[0]
        assert_error(is_valid_var(fieldName), '关键字必须是有效变量', lineno, line)
        pobjs = items[1] # [Prefix]objtyp[Subfix]
        refType = OBJ_REF_TYPE_MANAGE
        if pobjs[0] in ('@', '&'):
            refType = OBJ_REF_TYPE_REF
        if pobjs[0] == '^':
            needDump = False
        containerType = OBJ_CONTAINER_TYPE_SINGLE
        _isList = False
        if pobjs.endswith('[]'):
            _isList = True
            containerType = OBJ_CONTAINER_TYPE_LIST
        
        fieldType = pobjs.strip('[]').strip('@').strip('&').strip('^')
        if fieldType in TYPEMAPPING:
            containerType = OBJ_CONTAINER_TYPE_BASIC
            refType = OBJ_CONTAINER_TYPE_BASIC
            fieldType = TYPEMAPPING[fieldType] # rel gotype
            if _isList:
                fieldType = '[]' + fieldType
        
        return cls(lineno, line, fieldName, fieldType, containerType, refType, needDump, comment, parent)

class ObjModel(object):
    def __init__(self, root):
        self.root = root

    @classmethod
    def fromLines(cls, lines):
        startline = lineno = 1
        root = lines[0].strip()
        items = root.split(' ')
        assert_error(len(items) == 2, '无效的objmodel', lineno, root)
        assert_error(items[0] == 'root' and items[1] == 'object', '无效的objmodel', lineno, root)

        rootNode = ObjNode(lineno, root, None, None, None, None, False, None, None)
        deepParents = [rootNode]
        for line in lines[startline:]:
            lineno += 1

            if not line.strip():
                continue
            
            width = len(line) - len(line.lstrip())
            assert_error(width % TAB_WIDTH == 0, '宽度不是%s的倍数' % TAB_WIDTH, lineno, line)
            deep = int(width / TAB_WIDTH) # py2 vs py3
            assert_error(deep <= OBJ_MAX_DEEP_NUM, '深度不能超过%s' % OBJ_MAX_DEEP_NUM, lineno, line)

            line = line.strip()

            if deep == 0: # new top object
                # print 'add top object %s, %s' % (lineno, line)
                parentNode = deepParents[deep]
                node = ObjNode.fromLine(lineno, line, parentNode)
                if not node:
                    continue
                parentNode.addChildren(node)
                deepParents = [rootNode]
                deepParents.append(node)
            elif deep <= len(deepParents):
                parentNode = deepParents[deep]
                node = ObjNode.fromLine(lineno, line, parentNode)
                if not node:
                    continue
                # print 'add child %s to %s' % (node.fieldName, parentNode.fieldName)
                parentNode.addChildren(node)
                deepParents = deepParents[:deep+1]
                deepParents.append(node)
            else:
                assert_error(False, '缩进有问题 %s, 当前最大 %s' % (deep, len(deepParents)), lineno, line)

        return cls(rootNode)
    
    def checkObjCode(self):
        topObjects = []
        for child in self.root.children:
            topObjects.append(child.fieldType)
        
        for child in self.root.children:
            child.checkObjCode(topObjects)
        print('check done.')

    def genObjInitCode(self):
        code = '// Code generated by m2c. DO NOT EDIT.\npackage orm\n'
        code += 'import (\n'
        code += 'id "%s/ak/orm/id"\n' % MOD
        for child in self.root.children:
            assert isinstance(child, ObjNode)
            code += '%sObject "%s/ak/orm/%s"\n' % (child.fieldName, MOD, child.fieldName)
        code += ')\n'
        code += 'func Init(url string) {\n'
        code += 'mgoDB := getMGO(url)\n'
        code += 'id.Init(mgoDB)\n'
        for child in self.root.children:
            assert isinstance(child, ObjNode)
            code += '%sObject.Init(mgoDB)\n' % child.fieldName
        code += '}\n'
        code += '// DropDB Dangers!!!\n'
        code += 'func DropDB(url string) {\n'
        code += 'mgoDB := getMGO(url)\n'    
        code += 'mgoDB.DropDatabase()\n'
        code += '}\n'
        return code

    def genObjCode(self):
        # api model
        initPath = 'gopher/ak/orm/init.go'
        writeCode(initPath, self.genObjInitCode())

        # api group
        for child in self.root.children:
            assert isinstance(child, ObjNode)
            child.genObjCode()

        # api doc
        os.system('gofmt -w gopher/ak/orm')

    def genObjCodeV4(self):
        # db model
        for child in self.root.children:
            assert isinstance(child, ObjNode)
            child.genObjCodeV4()

        # api doc
        os.system('gofmt -w %s/model' % MOD)



#!/usr/bin/env python
#coding: utf-8
#by yangbin at 2018.11.28
from __future__ import print_function
vlogs = {
    '0.0.8': 'add text search template',
    '0.0.9': 'update service template url; auto check version',
    '0.1.0': 'add go source file wathdog',
    '0.1.1': 'add api version',
    '0.1.2': 'fix page',
    '0.1.3': 'add string type or/like template, fix without manager query bug',
    '0.1.4': 'update watchgo/orm',
    '0.1.5': 'update swagger',
    '0.1.6': 'fix manager list relation',
    '0.1.7': 'fix with string like',
    '0.1.8': 'page 1 base: 注意当前版本分页功能从1开始, 应用层不需要再做pageIndex-1 操作!!!', # add change_log notify
    '0.1.9': 'add batch update',
    '0.1.10': 'fix batch update',
    '0.1.11': 'add reqRaw/respRaw api; float/array type',
    '0.1.12': 'gop v0.0.3',
    '0.1.13': 'gop v0.0.3',
    '0.1.14': 'add orm $in/$nin, nest api param',
    '0.1.15': 'fix remove pdb',
    '0.1.16': 'export orm.GetQuery, orm.Scan filter empty object',
    '0.1.17': 'add api alias',
    '0.1.18': 'update nest apimodel',
    '0.1.19': 'update swagger',
    '0.2.0': 'add api param validator',
    '0.2.9': 'openApi v3 and apiResp model',
    '0.2.10': 'go mod vendor auto',
    '0.2.11': 'update orm AllResultTo',
    '0.2.12': 'add form tag for apimodel',
    '0.3.0': 'add new struct v4',
    '0.3.1': 'auto load .env',
    '0.3.2': 'fix load .env',
    '0.3.3': 'add any router',
    '0.3.4': 'add python3 support',
    '0.3.5': 'fix swagger tags/servers',
    '0.3.6': 'add service code template',
    '0.3.7': 'add cmdmodel',
    '0.3.8': 'add wildcard router',
    '0.3.9': 'add swagger codegen',
    '0.4.0': 'add swagger security, json resp model',
    '0.4.1': 'add m2c init from git repo',
    '0.4.2': 'fix pack',
    '0.4.3': 'add adhoc swaggerHtmlDoc cmd'
}

name = 'm2c'
version = '0.4.3'

from .conf import M2C_PATH, Color
from .helper import m2c_version

def change_log():
    curv = m2c_version()
    if not curv:
        return

    if curv == version:
        return

    print(Color.red(u'当前项目使用的m2c版本: %s' % curv))
    print(Color.red(u'使用m2c apicode 或者m2c objcode 可以更新当前版本'))
    print(Color.red(u'注意以下更新:'))
    for v, msg in vlogs.items():
        if v > curv:
            print(v, msg)


def check_version():
    return
    # PyPI's XMLRPC API is currently disabled due to unmanageable load and will be deprecated in the near future.
    # import os
    # import random
    # from .conf import Color
    # if random.randint(0, 10) % 2 == 0:
    #     return
    # # random check for speed up
    # change_log()
    # text = os.popen('pip search %s' % name).read()
    # need_update = False
    # for line in text.splitlines():
    #     print(line)
    #     line = line.strip()
    #     if line.startswith('LATEST:'):
    #         need_update = True
    # if need_update:
    #     print(Color.red('run: pip install -U %s, update it!' % name))
